#!/usr/bin/python3
import json
import unicodedata
import shutil
import sys
from pathlib import Path
from PyQt5 import QtCore, QtGui, QtWidgets, QtTest, uic

try:
    from ..ui import hsp_ui
    from ..__id__ import ID
    from ..backend import database, hsp, common
except ValueError:
    from __id__ import ID
    from backend import database, hsp, common

SET_FILE = Path.home() / ".config" / ID / "settings.json"
SLV_FILE = Path.home() / ".config" / ID / "hsp_solvents.json"
GRP_FILE = Path.home() / ".config" / ID / "hsp_groups.json"
LOCAL = str(Path(__file__).parents[0])


class InputLine(QtWidgets.QLineEdit):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.last = ""
        self.setTextMargins(5, 2, 5, 2)
        self.textChanged.connect(self._resetPalette)
        self.textChanged.connect(self._saveLast)
        self.editingFinished.connect(self._validate)
        self.setPlaceholderText("...")

        completer = common.Completer(self, list(parent.db.keys()))
        completer.activated.connect(parent._updateTarget)
        self.setCompleter(completer)

    def _saveLast(self, text):
        # Avoid selectAll when focus is lost to completer's popup
        self.last = text

    def focusInEvent(self, event):
        if self.last == self.text():
            QtCore.QTimer.singleShot(0, self.selectAll)
        QtWidgets.QLineEdit.focusInEvent(self, event)

    def keyPressEvent(self, event):
        if event.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
            self.parent.searchDual()
        QtWidgets.QLineEdit.keyPressEvent(self, event)

    def _errorPalette(self):
        fg = QtGui.QColor("#ffc7ce")
        bg = QtGui.QColor("#9c0006")
        palette = QtGui.QPalette()
        palette.setColor(QtGui.QPalette.Base, bg)
        palette.setColor(QtGui.QPalette.Text, fg)
        self.setPalette(palette)

    def _normalize(self, text):
        # Normalize case and diacritics
        normalized = unicodedata.normalize("NFKD", text)
        normalized = "".join(c for c in normalized if unicodedata.category(c) != "Mn")
        return normalized.lower()

    def _resetPalette(self, text):
        self.setPalette(self.parent.palette())
        self.setAttribute(QtCore.Qt.WA_SetPalette, False)

    def _validate(self):
        text = self.text()
        bland_names = {}
        for name in self.parent.db:
            n = self._normalize(name)
            bland_names[n] = name

        if text:
            bland_input = self._normalize(text)
            if bland_input in bland_names:
                valid_input = bland_names[bland_input]
                self.setText(valid_input)
            else:
                self._errorPalette()


class GroupTree(QtWidgets.QTreeWidget):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.current = "preferred"
        self.selectedItem = None
        self.setIndentation(5)
        self.setAlternatingRowColors(True)
        self.setStyleSheet("QTreeView::item {padding: 0 20px;}")
        self.setHeaderLabels(["Name", "𝛿D", "𝛿P", "𝛿H", "Molar Volume"])
        self.itemChanged.connect(self._itemChanged)
        self.itemClicked.connect(self._itemClicked)
        self.load()

    def keyPressEvent(self, event):
        item = self.currentItem()
        key = event.key()
        if key == QtCore.Qt.Key_Delete:
            index = self.indexOfTopLevelItem(item)
            self._deletePrompt(item.text(0), index)
        elif key == QtCore.Qt.Key_Space:
            if item:
                self._itemClicked(item)
        else:
            QtWidgets.QTreeWidget.keyPressEvent(self, event)

    def _delete(self, name, index):
        self.takeTopLevelItem(index)
        for grp in dict(self.parent.groups):
            if name in self.parent.groups[grp]:
                self.parent.groups[grp].remove(name)
        del self.parent.db[name]
        print(f"Deleted {name} from solvents database")

    def _deletePrompt(self, name, index):
        msg = common.ConfirmPrompt()
        msg.setWindowTitle("Delete confirmation")
        msg.setText(f"Please confirm deletion of '{name}'")
        if msg.exec_() == QtWidgets.QMessageBox.Apply:
            self._delete(name, index)

    def _itemChanged(self, item):
        state = bool(item.checkState(0))
        slv = item.text(0)
        db = self.parent.groups[self.current]
        if state:
            if slv not in db:
                db.append(slv)
        else:
            if slv in db:
                db.remove(slv)

    def _itemClicked(self, item):
        if item == self.selectedItem:
            state = bool(item.checkState(0))
            if state:
                item.setCheckState(0, QtCore.Qt.Unchecked)
            else:
                item.setCheckState(0, QtCore.Qt.Checked)
        else:
            self.selectedItem = item

    def load(self):
        self.clear()
        for slv in self.parent.db:
            item = GroupTreeItem(self, slv)
            if slv not in self.parent.groups[self.current]:
                item.setCheckState(0, QtCore.Qt.Unchecked)
                self.addTopLevelItem(item)

        group = self.parent.groups[self.current]
        for slv in sorted(group, reverse=True):
            item = GroupTreeItem(self, slv)
            item.setCheckState(0, QtCore.Qt.Checked)
            self.insertTopLevelItem(0, item)

        for i in range(0, self.columnCount()+1):
            self.resizeColumnToContents(i)

    def select(self, name):
        itemCount = self.topLevelItemCount()
        for item in range(itemCount):
            item = self.topLevelItem(item)
            if item.text(0) == name:
                index = self.indexOfTopLevelItem(item)
                item = self.takeTopLevelItem(index)
                self.insertTopLevelItem(0, item)
                self.setCurrentItem(item)
                self.selectedItem = item
                break


class GroupTreeItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, parent, name):
        super().__init__()
        self.setFlags(self.flags() &~ QtCore.Qt.ItemIsDropEnabled)
        self.setFlags(self.flags() &~ QtCore.Qt.ItemIsUserCheckable)

        slv = parent.parent.db[name]
        self.setText(0, name)
        self.setText(1, str(slv["dD"]))
        self.setText(2, str(slv["dP"]))
        self.setText(3, str(slv["dH"]))
        self.setText(4, str(slv["mv"]))


class SearchTree(QtWidgets.QTreeWidget):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.clipboard = parent.parent.clipboard()
        self.setAlternatingRowColors(True)
        self.setSortingEnabled(True)
        self.setIndentation(5)
        self.setStyleSheet("QTreeView::item { padding: 0 20px; }")
        self.setHeaderLabels(["Solvent 1", "Solvent 2", "Solvent 3", "Ratio", "𝛿D", "𝛿P", "𝛿H", "Molar Volume", "Ra"])
        self.sortByColumn(8, QtCore.Qt.AscendingOrder)
        self.header().setMinimumSectionSize(0)

    def keyPressEvent(self, event):
        ctrl = event.modifiers() == QtCore.Qt.ControlModifier
        c = event.key() == QtCore.Qt.Key_C
        if ctrl and c:
            self._clipboardExport()
        else:
            QtWidgets.QTreeWidget.keyPressEvent(self, event)

    def _clipboardExport(self):
        item = self.currentItem()
        if item:
            export = ""
            for i in range(0, self.columnCount()+1):
                if item.text(i):
                    export += f"{item.text(i)}\t"
            self.clipboard.setText(export)

    def addItem(self, item):
        names = (item.text(0), item.text(1))
        s1 = self.parent.ui.s1Line.text()
        s2 = self.parent.ui.s2Line.text()
        s1_valid = not s1 or s1 not in names
        s2_valid = not s2 or s2 not in names
        if s1_valid and s2_valid:
            self.addTopLevelItem(item)

    def resizeToContent(self):
        for i in range(0, self.columnCount()+1):
            self.resizeColumnToContents(i)

        if not self.parent.worker.hsp.all:
            self.setColumnWidth(1, 0)  # Solvent 2
            self.setColumnWidth(2, 0)  # Solvent 3
            self.setColumnWidth(3, 0)  # Ratio
        else:
            single = self.parent.ui.filterSingleBox.isChecked()
            dual = self.parent.ui.filterDualBox.isChecked()
            if single:
                self.setColumnWidth(1, 0)  # Solvent 2
                self.setColumnWidth(2, 0)  # Solvent 3
                self.setColumnWidth(3, 0)  # Ratio
            elif dual:
                self.setColumnWidth(2, 0)  # Solvent 3
            self.setColumnWidth(7, 0)  # Molar Volume


class SearchTreeItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.setFlags(self.flags() &~ QtCore.Qt.ItemIsDropEnabled)

    def __lt__(self, other):
        column = self.treeWidget().sortColumn()
        key1 = self.text(column)
        key2 = other.text(column)
        try:
            return float(key1) < float(key2)
        except ValueError:
            return key1.lower() < key2.lower()

    def setColor(self, column, color):
        try:
            if color == "none":
                self.setData(column, QtCore.Qt.BackgroundRole, None)
                self.setData(column, QtCore.Qt.ForegroundRole, None)
            elif color == "preferred":
                self.setForeground(column, QtGui.QColor("#004000"))
                self.setBackground(column, QtGui.QColor("#c6efce"))
            elif color == "usable":
                self.setForeground(column, QtGui.QColor("#553400"))
                self.setBackground(column, QtGui.QColor("#ffeb9c"))
            elif color == "hazardous":
                self.setForeground(column, QtGui.QColor("#9c0006"))
                self.setBackground(column, QtGui.QColor("#ffc7ce"))
        except RuntimeError:
            pass


class WorkerThread(QtCore.QObject):
    addItem = QtCore.pyqtSignal(object)
    working = QtCore.pyqtSignal(bool)
    resize = QtCore.pyqtSignal()

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.busy = False
        self.hsp = hsp.HSP()

    def _makeItem(self, slv):
        slv = self.hsp.all[slv]
        item = SearchTreeItem(self)
        item.setText(0, slv["s1"])

        if "s3" in slv:
            r1, r2, r3 = slv['ratio']
            ratio = f"{r1}:{r2}:{r3}"
            item.setText(1, slv["s2"])
            item.setText(2, slv["s3"])
            item.setText(3, ratio)

        elif "s2" in slv:
            r1, r2 = slv['ratio']
            ratio = f"{r1}:{r2}"
            item.setText(1, slv["s2"])
            item.setText(3, ratio)

        item.setText(4, str(slv["dD"]))
        item.setText(5, str(slv["dP"]))
        item.setText(6, str(slv["dH"]))
        item.setText(8, str(slv["Ra"]))
        item = self._setColor(item)
        return item

    def _setColor(self, item):
        for grp in ("preferred", "usable", "hazardous"):
            for col in range (0, 3):
                if item.text(col) in self.parent.groups[grp]:
                    item.setColor(col, grp)
        return item

    def _show(self, db):
        for slv in db:
            item = self._makeItem(slv)
            self.addItem.emit(item)
        self.resize.emit()

    @QtCore.pyqtSlot(str, tuple, dict, list)
    def search(self, scope, dph, target, groups):
        self.busy = True
        self.working.emit(True)
        self.hsp.search(scope, dph, target, groups)
        self.working.emit(False)
        self.busy = False

    @QtCore.pyqtSlot()
    def showAll(self):
        self._show(self.hsp.all)

    @QtCore.pyqtSlot()
    def showSingle(self):
        self._show(self.hsp.single)

    @QtCore.pyqtSlot()
    def showDual(self):
        self._show(self.hsp.dual)

    @QtCore.pyqtSlot()
    def showTriple(self):
        self._show(self.hsp.triple)

    @QtCore.pyqtSlot()
    def stop(self):
        self.hsp.abort = True


class Main(QtWidgets.QMainWindow):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self._loadDatabases()
        self._uiInit()
        self.show()
        self.showMaximized()

        self.timer = QtCore.QElapsedTimer()
        self.worker = WorkerThread(self)
        self.workerThread = QtCore.QThread()
        self.worker.moveToThread(self.workerThread)
        self.workerThread.start()
        self.worker.addItem.connect(self.ui.searchTree.addItem)
        self.worker.working.connect(self._updateStatus)
        self.worker.working.connect(self._searchBlock)
        self.worker.resize.connect(self.ui.searchTree.resizeToContent)

        self.ui.searchAllButton.clicked.connect(self.searchAll)
        self.ui.searchSingleButton.clicked.connect(self.searchSingle)
        self.ui.searchDualButton.clicked.connect(self.searchDual)
        self.ui.searchTripleButton.clicked.connect(self.searchTriple)
        self.ui.filterAllBox.clicked.connect(self._updateFilter)
        self.ui.filterSingleBox.clicked.connect(self._updateFilter)
        self.ui.filterDualBox.clicked.connect(self._updateFilter)
        self.ui.filterTripleBox.clicked.connect(self._updateFilter)
        self.ui.groupPreferedButton.clicked.connect(self._updateGroupView)
        self.ui.groupUsableButton.clicked.connect(self._updateGroupView)
        self.ui.groupHazardousButton.clicked.connect(self._updateGroupView)
        self.ui.groupCustomButton.clicked.connect(self._updateGroupView)
        self.ui.groupClearButton.clicked.connect(self._groupClear)
        self.ui.groupAddButton.clicked.connect(self._groupAdd)
        self.ui.resetButton.clicked.connect(self._resetPrompt)

        self.ui.searchTree.currentItemChanged.connect(self._itemChanged)
        self.ui.s1Line.editingFinished.connect(self._updateTarget)
        self.ui.s2Line.editingFinished.connect(self._updateTarget)
        self.ui.dTargetBox.valueChanged.connect(self._customTarget)
        self.ui.pTargetBox.valueChanged.connect(self._customTarget)
        self.ui.hTargetBox.valueChanged.connect(self._customTarget)
        self.ui.targetS1VolBox.valueChanged.connect(self._updateVolume)
        self.ui.targetS2VolBox.valueChanged.connect(self._updateVolume)

        self.ui.dTargetBox.installEventFilter(self)
        self.ui.pTargetBox.installEventFilter(self)
        self.ui.hTargetBox.installEventFilter(self)
        self.ui.dGroupBox.installEventFilter(self)
        self.ui.pGroupBox.installEventFilter(self)
        self.ui.hGroupBox.installEventFilter(self)
        self.ui.mvGroupBox.installEventFilter(self)

        self._settingsInit()
        self._settingsLoad()
        self._updateTarget()
        self._showDefault()

    def closeEvent(self, event):
        self._solventsSave()
        self._groupsSave()
        self._settingsSave()

    def eventFilter(self, obj, event):
        # Select all on focus
        if event.type() == QtCore.QEvent.FocusIn:
            QtCore.QTimer.singleShot(0, obj.selectAll)

        # Replace , with .
        if event.type() == QtCore.QEvent.KeyPress:
            if event.key() == QtCore.Qt.Key_Comma:
                QtTest.QTest.keyPress(obj, QtCore.Qt.Key_Period, QtCore.Qt.NoModifier)
                QtTest.QTest.keyRelease(obj, QtCore.Qt.Key_Period, QtCore.Qt.NoModifier)
                return True
        return QtCore.QObject.event(obj, event)

    def _blockTargetBoxes(self, state):
        # Qt limitation; signal is blocked to ignore programatically changed value in QSpinBox
        self.ui.dTargetBox.blockSignals(state)
        self.ui.pTargetBox.blockSignals(state)
        self.ui.hTargetBox.blockSignals(state)

    def _customTarget(self):
        self.worker.hsp.reset()
        self.ui.s1Line.clear()
        self.ui.s2Line.clear()

    def _filter(self):
        if self.ui.filterAllBox.isChecked():
            return "showAll"
        elif self.ui.filterSingleBox.isChecked():
            return "showSingle"
        elif self.ui.filterDualBox.isChecked():
            return "showDual"
        elif self.ui.filterTripleBox.isChecked():
            return "showTriple"

    def _groupAdd(self):
        name = self.ui.groupNameLine.text()
        if name:
            dD = self.ui.dGroupBox.value()
            dP = self.ui.pGroupBox.value()
            dH = self.ui.hGroupBox.value()
            mv = self.ui.mvGroupBox.value()
            self.db[name] = {"dD": dD, "dP": dP, "dH": dH, "mv": mv}
            self._solventsSave()
            print(f"Added {name} to solvents database")

            self.ui.groupTree.load()
            self.ui.groupTree.select(name)
            self.ui.groupTree.setFocus(True)
            self._groupClear()

    def _groupClear(self):
        self.ui.groupNameLine.clear()
        self.ui.dGroupBox.setValue(0)
        self.ui.pGroupBox.setValue(0)
        self.ui.hGroupBox.setValue(0)
        self.ui.mvGroupBox.setValue(0)

    def _groupsSave(self):
        with open(GRP_FILE, "w") as f:
            f.write(json.dumps(self.groups, indent=2, sort_keys=False))
        print(f"Saved groups database")

    def _itemChanged(self, current, previous):
        if current:
            self.ui.dSelectLine.setText(current.text(4))
            self.ui.pSelectLine.setText(current.text(5))
            self.ui.hSelectLine.setText(current.text(6))
            self._updateDelta()

    def _loadDatabases(self):
        slv = Path(SLV_FILE).is_file()
        grp = Path(GRP_FILE).is_file()

        if slv and grp:
            with open(SLV_FILE, "r") as f:
                self.db = json.load(f)
            with open(GRP_FILE, "r") as f:
                self.groups = json.load(f)
        else:
            cfg = Path.home() / ".config" / ID
            Path(cfg).mkdir(parents=True, exist_ok=True)
            self._reset()
            print("Created default data files")

    def _readEnableDPH(self):
        dD = self.ui.dTargetEnableBox.isChecked()
        dP = self.ui.pTargetEnableBox.isChecked()
        dH = self.ui.hTargetEnableBox.isChecked()
        return (dD, dP, dH)

    def _readGroups(self):
        groups = []
        if self.ui.groupPreferedBox.isChecked():
            groups.append("preferred")
        if self.ui.groupUsableBox.isChecked():
            groups.append("usable")
        if self.ui.groupHazardousBox.isChecked():
            groups.append("hazardous")
        if self.ui.groupCustomBox.isChecked():
            groups.append("custom")
        return groups

    def _readTarget(self):
        self._updateTarget()
        dD = self.ui.dTargetBox.value()
        dP = self.ui.pTargetBox.value()
        dH = self.ui.hTargetBox.value()
        return {"dD": dD, "dP": dP, "dH": dH}

    def _reset(self):
        slv_default = Path(LOCAL).parents[0] / "data" / "hsp_solvents.json"
        grp_default = Path(LOCAL).parents[0] / "data" / "hsp_groups.json"
        shutil.copyfile(slv_default, SLV_FILE)
        shutil.copyfile(grp_default, GRP_FILE)
        self._loadDatabases()

    def _resetPrompt(self):
        msg = common.ConfirmPrompt()
        msg.setWindowTitle("Reset confirmation")
        msg.setText("This will reset solvent database and all\n" +
                    "groups content to their default value.\n" +
                    "\nDo you want to proceed?")
        if msg.exec_() == QtWidgets.QMessageBox.Apply:
            self._reset()
            self.ui.groupTree.load()

    def _search(self, scope):
        if self.worker.busy:
            QtCore.QMetaObject.invokeMethod(self.worker, "stop", QtCore.Qt.DirectConnection)
        else:
            target = self._readTarget()
            self.ui.searchTree.clear()
            self.ui.dTargetLine.setText("%.1f" % target["dD"])
            self.ui.pTargetLine.setText("%.1f" % target["dP"])
            self.ui.hTargetLine.setText("%.1f" % target["dH"])
            target = QtCore.Q_ARG(dict, target)
            scope = QtCore.Q_ARG(str, scope)
            groups = QtCore.Q_ARG(list, self._readGroups())
            dph = QtCore.Q_ARG(tuple, self._readEnableDPH())

            QtCore.QMetaObject.invokeMethod(self.worker, "search", QtCore.Qt.QueuedConnection, scope, dph, target, groups)
            QtCore.QMetaObject.invokeMethod(self.worker, self._filter(), QtCore.Qt.QueuedConnection)

    def _searchBlock(self, state):
        state = not state
        if state:
            self.ui.searchAllButton.setText("Optimize")
        else:
            self.ui.searchAllButton.setText("Stop")
        self.ui.searchSingleButton.setEnabled(state)
        self.ui.searchDualButton.setEnabled(state)
        self.ui.searchTripleButton.setEnabled(state)

    def _setColor(self, item):
        for grp in ("preferred", "usable", "hazardous"):
            for col in range (0, 3):
                if item.text(col) in self.groups[grp]:
                    item.setColor(col, grp)
        return item

    def _settingsInit(self):
        self.settings = database.Database(SET_FILE)
        self.settings.fields = [self.ui.s1Line,
                                self.ui.s2Line,
                                self.ui.targetS1VolBox,
                                self.ui.targetS2VolBox,
                                self.ui.dTargetBox,
                                self.ui.pTargetBox,
                                self.ui.hTargetBox,
                                self.ui.dTargetEnableBox,
                                self.ui.pTargetEnableBox,
                                self.ui.hTargetEnableBox,
                                self.ui.groupPreferedBox,
                                self.ui.groupUsableBox,
                                self.ui.groupHazardousBox,
                                self.ui.groupCustomBox,
                                self.ui.filterAllBox,
                                self.ui.filterSingleBox,
                                self.ui.filterDualBox,
                                self.ui.filterTripleBox]

    def _settingsLoad(self):
        db = self.settings.db["hsp"]
        self._blockTargetBoxes(True)
        self.settings.load(db)
        self._blockTargetBoxes(False)

    def _settingsSave(self):
        db = self.settings.db["hsp"]
        self.settings.save(db)

    def _showDefault(self):
        for slv in self.db:
            item = SearchTreeItem(self)
            item.setText(0, slv)
            item.setText(4, str(self.db[slv]["dD"]))
            item.setText(5, str(self.db[slv]["dP"]))
            item.setText(6, str(self.db[slv]["dH"]))
            item.setText(7, str(self.db[slv]["mv"]))
            item.setText(8, "-")
            item = self._setColor(item)
            self.ui.searchTree.addTopLevelItem(item)
        self.ui.searchTree.resizeToContent()

    def _solventsSave(self):
        with open(SLV_FILE, "w") as f:
            f.write(json.dumps(self.db, indent=2, sort_keys=False))
        print(f"Saved solvents database")

    def _uiInit(self):
        try:
            self.ui = hsp_ui.Ui_MainWindow()
            self.ui.setupUi(self)
        except NameError:
            path = Path(LOCAL).parents[0] / "ui" / "hsp_ui.ui"
            self.ui = uic.loadUi(path, self)

        path = str(Path(LOCAL).parents[0] / "ui" / "icons" / "qtlab-hsp.svg")
        icon = QtGui.QIcon(path)
        self.setWindowIcon(icon)
        self.setWindowTitle("HSP Utility - Hansen Solubility Parameters")

        self.ui.searchTree = SearchTree(self)
        self.ui.groupTree = GroupTree(self)
        self.s1Line = InputLine(self)
        self.s2Line = InputLine(self)
        self.s1Line.setObjectName("s1Line")
        self.s2Line.setObjectName("s2Line")
        self.ui.searchTreeLayout.insertWidget(0, self.ui.searchTree)
        self.ui.groupTreeLayout.insertWidget(0, self.ui.groupTree)
        self.ui.inputLayout.insertWidget(0, self.s1Line)
        self.ui.inputLayout.insertWidget(1, self.s2Line)

    def _updateDelta(self):
        try:
            dD = float(self.ui.dSelectLine.text()) - float(self.ui.dTargetLine.text())
            dP = float(self.ui.pSelectLine.text()) - float(self.ui.pTargetLine.text())
            dH = float(self.ui.hSelectLine.text()) - float(self.ui.hTargetLine.text())
            self.ui.dDeltaLine.setText("%.1f" % dD)
            self.ui.pDeltaLine.setText("%.1f" % dP)
            self.ui.hDeltaLine.setText("%.1f" % dH)
        except ValueError:
            pass

    def _updateFilter(self):
        if self.worker.hsp.all:
            self.ui.searchTree.clear()
        QtCore.QMetaObject.invokeMethod(self.worker, self._filter(), QtCore.Qt.QueuedConnection)

    def _updateGroupView(self):
        old = self.ui.groupTree.current
        if self.ui.groupPreferedButton.isChecked():
            new = "preferred"
        elif self.ui.groupUsableButton.isChecked():
            new = "usable"
        elif self.ui.groupHazardousButton.isChecked():
            new = "hazardous"
        elif self.ui.groupCustomButton.isChecked():
            new = "custom"

        if old != new:
            self.ui.groupTree.current = new
            self.ui.groupTree.load()

    def _updateStatus(self, working):
        if working:
            self.timer.start()
            self.ui.statusBar().showMessage("Calculation in progress...")
        else:
            elapsed = round((self.timer.elapsed()/1000), 2)
            msg = f"Search done after {elapsed}s"
            self.ui.statusBar().showMessage(msg)

    def _updateTarget(self):
        self._blockTargetBoxes(True)
        s1 = self.ui.s1Line.text()
        s2 = self.ui.s2Line.text()
        s1_vol = self.ui.targetS1VolBox.value()
        s2_vol = self.ui.targetS2VolBox.value()

        slv = None
        if s1 in self.db and s2 in self.db:
            s1, s2 = hsp.dph(s1), hsp.dph(s2)
            s1["vol"], s2["vol"] = s1_vol, s2_vol
            slv = hsp.combine(s1, s2)
        elif s1 in self.db:
            slv = self.db[s1]

        if slv:
            self.ui.dTargetBox.setValue(slv["dD"])
            self.ui.pTargetBox.setValue(slv["dP"])
            self.ui.hTargetBox.setValue(slv["dH"])
            self._updateDelta()
        self._blockTargetBoxes(False)

    def _updateVolume(self, value):
        sender = QtCore.QObject.sender(self)
        value = 100 - value
        if self.ui.targetS1VolBox is sender:
            self.ui.targetS2VolBox.setValue(value)
        else:
            self.ui.targetS1VolBox.setValue(value)
        self._updateTarget()

    def searchAll(self):
        self.ui.filterAllBox.setChecked(True)
        self._search(scope="all")

    def searchSingle(self):
        self.ui.filterSingleBox.setChecked(True)
        self._search(scope="single")

    def searchDual(self):
        self.ui.filterDualBox.setChecked(True)
        self._search(scope="dual")

    def searchTriple(self):
        self.ui.filterTripleBox.setChecked(True)
        self._search(scope="triple")



VERSION = "0.0.22"

ID = "qtlab"
APP_NAME = "QtLab Utilities"
DEPENDANCIES = ["openpyxl", "pandas", "scipy"]
DESCRIPTION = ""
KEYWORDS = "csv calculator hsp chart excel converter chemistry biology science"
AUTHOR = "William Belanger"
URL = f"https://gitlab.com/william.belanger/{ID}"

CLASSIFIERS = [
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Programming Language :: Python :: 3.6",
    "Operating System :: POSIX :: Linux",
    "Operating System :: Microsoft :: Windows :: Windows 7",
]

SHORTCUTS = [
    ("calc", "Solutions Calculator"),
    ("hsp", "Solvent Designer (HSP Calculator)"),
    ("peak", "CSV to Chart Converter"),
]

HELP = [
    "--help, -h: Show this help message",
    "--hsp: Open the HSP utility",
    "--calc: Open the solution calculator",
    "   --toggle: Toggle the calculator on or off",
    "--peak: Open the csv to chart conversion utility",
    "--quit: Close the application",
]




#!/usr/bin/python3
DEFAULT = \
{
    "hsp": {
        "s1Line": "Acetonitrile",
        "s2Line": "",
        "targetS1VolBox": 100,
        "targetS2VolBox": 0,
        "dTargetBox": 0.0,
        "pTargetBox": 0.0,
        "hTargetBox": 0.0,
        "dTargetEnableBox": True,
        "pTargetEnableBox": True,
        "hTargetEnableBox": True,
        "groupPreferedBox": True,
        "groupUsableBox": False,
        "groupHazardousBox": False,
        "groupCustomBox": False,
        "filterAllBox": True,
        "filterSingleBox": False,
        "filterDualBox": False,
        "filterTripleBox": False,
    },
    "calc": {
        "molConcCombo": "millimolar (mM)",
        "molVolCombo": "milliliters (mL)",
        "dilC1Combo": "",
        "dilC2Combo": "",
        "dilV2Combo": "",
        "dilResultCombo": "milliliters (mL)",
        "notes": "μ°∙Å¢αβλΔδ™",
        "copyFormatLine": "TODO",
        "copyClipboardBox": True,
        "copyNotesBox": True,
        "escapeKeyBox": True,
        "displayFormulaBox": False,
    },
    "peak": {
        "autodetect": True,
        "last profile": "Cary WinUV (v3)",
        "profiles":
        {
            "Cary WinUV (v3)":
            {
                "focusBox": True,
                "statsBox": True,
                "focusMinBox": 500,
                "focusMaxBox": 800,
                "statsMinBox": 500,
                "statsMaxBox": 800,
                "xLine": "Wavelength (nm)",
                "yLine": "Absorbance",
                "titleStatsLine": "Sample",
                "xStatsLine": "λmax (nm)",
                "yStatsLine": "λmax (abs)",
                "output path": "",
                "input path": "",
            },
            "OMNIC for Dispertive Raman (v8)":
            {
                "focusBox": False,
                "statsBox": False,
                "focusMinBox": 0,
                "focusMaxBox": 0,
                "statsMinBox": 0,
                "statsMaxBox": 0,
                "xLine": "Raman shift (cm-1)",
                "yLine": "Intensity",
                "titleStatsLine": "Sample",
                "xStatsLine": "λmax (cm-1)",
                "yStatsLine": "λmax (intensity)",
                "output path": "",
                "input path": "",
            }
        }
    },
}




#!/usr/bin/python3
import json
import shutil
import sys
from pathlib import Path
from PyQt5 import QtCore, QtGui, QtWidgets, QtTest, uic

try:
    from ..ui import calc_ui
    from ..__id__ import ID
    from ..backend import database, calc, common
except ValueError:
    from __id__ import ID
    from backend import database, calc, common

SET_FILE = Path.home() / ".config" / ID / "settings.json"
SOLIDS_FILE = Path.home() / ".config" / ID / "calc_solids.json"
SOLUTIONS_FILE = Path.home() / ".config" / ID / "calc_solutions.json"
LOCAL = str(Path(__file__).parents[0])


class InputLine(QtWidgets.QLineEdit):
    def __init__(self, parent, db):
        super().__init__()
        self.parent = parent
        self.db = db
        self.last = ""
        self.setTextMargins(5, 2, 5, 2) ##pour tous
        self.textChanged.connect(self._saveLast)
        self.textChanged.connect(self._validate)
        self.setPlaceholderText("...")
        self.completerReload()

    def focusInEvent(self, event):
        if self.last == self.text():
            QtCore.QTimer.singleShot(0, self.selectAll)
        QtWidgets.QLineEdit.focusInEvent(self, event)

    def _completerActivated(self):
        self.setFocus(False)
        index = self.parent.ui.tabWidget.currentIndex()
        if index == 0:
            self.parent.ui.molConcBox.setFocus(True)
        elif index == 1:
            self.parent.ui.dilC2Box.setFocus(True)
        QtCore.QTimer.singleShot(0, self.deselect)

    def _saveLast(self, text):
        # Avoid selectAll when focus is lost to completer's popup
        self.last = text

    def _validate(self, text):
        if text in self.db:
            ui = self.parent.ui
            index = ui.tabWidget.currentIndex()
            value = self.db[text]
            if index == 0:
                ui.molMWBox.setValue(value)
            elif index == 1:
                ui.dilC1Box.setValue(value)

    def completerReload(self):
        keys = list(self.db.keys())
        completer = common.Completer(self, keys)
        completer.activated.connect(self._completerActivated)
        self.setCompleter(completer)


class FramelessWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        font = self.font()
        font.setBold(True)
        self.closeButton = QtWidgets.QToolButton(self)
        self.closeButton.setText("✖")
        self.closeButton.setFont(font)
        self.closeButton.setAutoRaise(True)
        self.closeButton.setFocusPolicy(QtCore.Qt.NoFocus)
        self.closeButton.setStyleSheet("margin-bottom: 3px;")
        self.closeButton.clicked.connect(self.close)

    def keyPressEvent(self, event):
        # Escape key close the window
        if event.key() == QtCore.Qt.Key_Escape:
            if self.ui.escapeKeyBox.isChecked():
                self.close()
        QtWidgets.QMainWindow.keyPressEvent(self, event)

    def mouseMoveEvent(self, event):
        # Enable mouse dragging
        if event.buttons() == QtCore.Qt.LeftButton:
            self.move(event.globalPos() - self.dragPosition)
            event.accept()

    def mousePressEvent(self, event):
        # Enable mouse dragging
        if event.button() == QtCore.Qt.LeftButton:
            self.dragPosition = event.globalPos() - self.frameGeometry().topLeft()
            event.accept()

    def paintEvent(self, event):
        # Draw a one pixel border
        borderColor = QtGui.QColor("black")
        bgColor = QtGui.QColor(self.palette().color(QtGui.QPalette.Background))
        painter = QtGui.QPainter(self)
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(QtGui.QBrush(borderColor))
        painter.drawRect(0, 0, self.width(), self.height())
        painter.setBrush(QtGui.QBrush(bgColor))
        painter.drawRect(1, 1, self.width()-2, self.height()-2)


class TabBar(QtWidgets.QTabBar):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.addTab("Molar solution")
        self.addTab("Solution dilution")
        self.addTab("Notes")
        self.addTab("Settings")
        self.setDocumentMode(True)
        self.currentChanged.connect(self._currentChanged)

    def _currentChanged(self, index):
        if index == 0:
            self.parent._calcMolar()
        elif index == 1:
            self.parent._calcDilution()
        else:
            self.parent.ui.statusLabel.clear()

    def mouseMoveEvent(self, event):
        # Enable mouse dragging
        if event.buttons() == QtCore.Qt.LeftButton:
            self.parent.move(event.globalPos() - self.dragPosition)
            event.accept()
        QtWidgets.QTabBar.mouseMoveEvent(self, event)

    def mousePressEvent(self, event):
        # Enable mouse dragging
        if event.button() == QtCore.Qt.LeftButton:
            self.dragPosition = event.globalPos() - self.parent.frameGeometry().topLeft()
            event.accept()
        QtWidgets.QTabBar.mousePressEvent(self, event)


class Main(FramelessWindow):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.calc = calc.Calc()
        self._uiInit()
        self._iconsInit()
        self._settingsInit()
        self._settingsLoad()
        self.show()

    def closeEvent(self, event):
        self._saveDatabases()
        self._settingsSave()
        QtWidgets.QMainWindow.closeEvent(self, event)

    def eventFilter(self, obj, event):
        # Select all on focus
        if event.type() == QtCore.QEvent.FocusIn:
            QtCore.QTimer.singleShot(0, obj.selectAll)

        # Replace , with .
        if event.type() == QtCore.QEvent.KeyPress:
            if event.key() == QtCore.Qt.Key_Comma:
                QtTest.QTest.keyPress(obj, QtCore.Qt.Key_Period, QtCore.Qt.NoModifier)
                QtTest.QTest.keyRelease(obj, QtCore.Qt.Key_Period, QtCore.Qt.NoModifier)
                return True
        return QtCore.QObject.event(obj, event)

    def _calcMolar(self):
        mw = self.ui.molMWBox.value()
        conc = self.ui.molConcBox.value()
        vol = self.ui.molVolBox.value()
        concUnit = self.ui.molConcCombo.currentText()
        volUnit = self.ui.molVolCombo.currentText()
        args = mw, conc, concUnit, vol, volUnit

        mass = self.calc.molar_solution(*args)
        self.ui.molResultBox.setValue(mass)

        formula = self.calc.molar_string(*args)
        self._updateFormula(formula)

    def _calcDilution(self):
        c1 = self.ui.dilC1Box.value()
        v2 = self.ui.dilV2Box.value()
        c2 = self.ui.dilC2Box.value()
        c1Unit = self.ui.dilC1Combo.currentText()
        v2Unit = self.ui.dilV2Combo.currentText()
        c2Unit = self.ui.dilC2Combo.currentText()
        resultUnit = self.ui.dilResultCombo.currentText()
        args = c1, c1Unit, c2, c2Unit, v2, v2Unit, resultUnit

        v1 = self.calc.dilution(*args)
        self.ui.dilResultBox.setValue(v1)

        formula = self.calc.dilution_string(*args)
        self._updateFormula(formula)

    def _dilClear(self):
        self.ui.dilNameLine.clear()
        self.ui.dilC1Box.setValue(0)
        self.ui.dilV2Box.setValue(0)
        self.ui.dilC2Box.setValue(0)

    def _dilDbAdd(self):
        name = self.ui.dilNameLine.text()
        unit = self.ui.dilC1Combo.currentText()
        c1 = self.ui.dilC1Box.value()
        c1_molar = c1 / calc.units_val[unit]
        self.dbSolutions[name] = c1_molar
        self.ui.dilNameLine.db = self.dbSolutions
        self.ui.dilNameLine.completerReload()

    def _dilDbDel(self):
        name = self.ui.dilNameLine.text()
        if name in self.dbSolutions:
            del self.dbSolutions[name]
            print(f"Deleted {name} from solutions database")
        self.ui.dilNameLine.clear()
        self.ui.dilC1Box.setValue(0)
        self.ui.dilNameLine.db = self.dbSolutions
        self.ui.dilNameLine.completerReload()

    def _iconsInit(self):
        icons = {}
        iconsPath = Path(__file__).parents[1] / "ui" / "icons"
        files = [x for x in iconsPath.glob("*.svg") if x.is_file()]
        for path in files:
            icons[path.stem] = QtGui.QIcon(str(path))
        self.ui.molDbAddButton.setIcon(icons["dbAdd"])
        self.ui.molDbDelButton.setIcon(icons["dbDel"])
        self.ui.molClearButton.setIcon(icons["clear"])
        self.ui.molCopyButton.setIcon(icons["copy"])
        self.ui.dilDbAddButton.setIcon(icons["dbAdd"])
        self.ui.dilDbDelButton.setIcon(icons["dbDel"])
        self.ui.dilClearButton.setIcon(icons["clear"])

    def _molCopy(self):
        name = self.ui.molNameLine.text()
        conc = self.ui.molConcBox.value()
        vf = self.ui.molVolBox.value()
        mass = self.ui.molResultBox.value()

        vfUnit = self.ui.molVolCombo.currentText()
        concUnit = self.ui.molConcCombo.currentText()
        vfUnit = calc.units_str[vfUnit]
        concUnit = calc.units_str[concUnit]

        if name and mass:
            # NaCl 200mM 5g/12mL
            txt = f"{name}\t{conc}{concUnit}\t{mass}g / {vf}{vfUnit}"
            if self.ui.copyClipboardBox.isChecked():
                self.parent.clipboard().setText(txt)
            if self.ui.copyNotesBox.isChecked():
                self.ui.notes.appendPlainText(txt)

    def _loadDatabases(self):
        solids = Path(SOLIDS_FILE ).is_file()
        solutions = Path(SOLUTIONS_FILE).is_file()

        if solids and solutions:
            with open(SOLIDS_FILE , "r") as f:
                self.dbSolids = json.load(f)
            with open(SOLUTIONS_FILE, "r") as f:
                self.dbSolutions = json.load(f)
        else:
            cfg = Path.home() / ".config" / ID
            Path(cfg).mkdir(parents=True, exist_ok=True)
            self._reset()
            print("Created default data files")

    def _saveDatabases(self):
        with open(SOLIDS_FILE, "w") as f:
            f.write(json.dumps(self.dbSolids, indent=2, sort_keys=False))
        with open(SOLUTIONS_FILE, "w") as f:
            f.write(json.dumps(self.dbSolutions, indent=2, sort_keys=False))

    def _molClear(self):
        self.ui.molNameLine.clear()
        self.ui.molMWBox.setValue(0)
        self.ui.molConcBox.setValue(0)
        self.ui.molVolBox.setValue(0)

    def _molDbAdd(self):
        name = self.ui.molNameLine.text()
        mw = self.ui.molMWBox.value()
        self.dbSolids[name] = mw
        self.ui.molNameLine.db = self.dbSolids
        self.ui.molNameLine.completerReload()

    def _molDbDel(self):
        name = self.ui.molNameLine.text()
        if name in self.dbSolids:
            del self.dbSolids[name]
            print(f"Deleted {name} from compounds database")
        self.ui.molNameLine.clear()
        self.ui.molMWBox.setValue(0)
        self.ui.molNameLine.db = self.dbSolids
        self.ui.molNameLine.completerReload()

    def _reset(self):
        solids_default = Path(LOCAL).parents[0] / "data" / "calc_solids.json"
        solutions_default = Path(LOCAL).parents[0] / "data" / "calc_solutions.json"
        shutil.copyfile(solids_default, SOLIDS_FILE)
        shutil.copyfile(solutions_default, SOLUTIONS_FILE)
        self._loadDatabases()

    def _resetDbs(self):
        self.ui.dilNameLine.db = self.dbSolutions
        self.ui.molNameLine.db = self.dbSolids
        self.ui.dilNameLine.completerReload()
        self.ui.molNameLine.completerReload()

    def _resetPrompt(self):
        msg = common.ConfirmPrompt()
        msg.setWindowTitle("Reset confirmation")
        msg.setText("This will reset databases of all compounds\n" +
                    "and solutions to their default value.\n" +
                    "\nDo you want to proceed?")
        if msg.exec_() == QtWidgets.QMessageBox.Apply:
            self._reset()
            self._resetDbs()

    def _settingsInit(self):
        self.settings = database.Database(SET_FILE)
        self.settings.fields = [self.ui.molConcCombo,
                                self.ui.molVolCombo,
                                self.ui.dilC1Combo,
                                self.ui.dilV2Combo,
                                self.ui.dilC2Combo,
                                self.ui.dilResultCombo,
                                self.ui.copyClipboardBox,
                                self.ui.copyNotesBox,
                                self.ui.displayFormulaBox,
                                self.ui.escapeKeyBox,
                                self.ui.notes]

    def _settingsLoad(self):
        db = self.settings.db["calc"]
        self.settings.load(db)

    def _settingsSave(self):
        db = self.settings.db["calc"]
        self.settings.save(db)

    def _uiInit(self):
        try:
            self.ui = calc_ui.Ui_MainWindow()
            self.ui.setupUi(self)
        except NameError:
            path = Path(LOCAL).parents[0] / "ui" / "calc_ui.ui"
            self.ui = uic.loadUi(path, self)

        path = str(Path(LOCAL).parents[0] / "ui" / "icons" / "qtlab-calc.svg")
        icon = QtGui.QIcon(path)
        self.setWindowIcon(icon)
        self.setWindowTitle("Solutions Calculator")
        self.setWindowFlags(QtCore.Qt.FramelessWindowHint | QtCore.Qt.WindowStaysOnTopHint)
        self._loadDatabases()

        self.tabBar = TabBar(self)
        self.ui.tabWidget.setTabBar(self.tabBar)
        self.ui.tabWidget.setCornerWidget(self.closeButton)
        self.ui.resetButton.clicked.connect(self._resetPrompt)

        self.molNameLine = InputLine(self, db=self.dbSolids)
        self.ui.molInputLayout.insertWidget(0, self.molNameLine)
        self.ui.molMWBox.valueChanged.connect(self._calcMolar)
        self.ui.molConcBox.valueChanged.connect(self._calcMolar)
        self.ui.molVolBox.valueChanged.connect(self._calcMolar)
        self.ui.molConcCombo.currentTextChanged.connect(self._calcMolar)
        self.ui.molVolCombo.currentTextChanged.connect(self._calcMolar)
        self.ui.molDbAddButton.clicked.connect(self._molDbAdd)
        self.ui.molDbDelButton.clicked.connect(self._molDbDel)
        self.ui.molClearButton.clicked.connect(self._molClear)
        self.ui.molCopyButton.clicked.connect(self._molCopy)

        self.dilNameLine = InputLine(self, db=self.dbSolutions)
        self.ui.dilInputLayout.insertWidget(0, self.dilNameLine)
        self.ui.dilC1Box.valueChanged.connect(self._calcDilution)
        self.ui.dilV2Box.valueChanged.connect(self._calcDilution)
        self.ui.dilC2Box.valueChanged.connect(self._calcDilution)
        self.ui.dilC1Combo.currentTextChanged.connect(self._calcDilution)
        self.ui.dilC2Combo.currentTextChanged.connect(self._calcDilution)
        self.ui.dilV2Combo.currentTextChanged.connect(self._calcDilution)
        self.ui.dilResultCombo.currentTextChanged.connect(self._calcDilution)
        self.ui.dilDbAddButton.clicked.connect(self._dilDbAdd)
        self.ui.dilDbDelButton.clicked.connect(self._dilDbDel)
        self.ui.dilClearButton.clicked.connect(self._dilClear)

        self.ui.molMWBox.installEventFilter(self)
        self.ui.molConcBox.installEventFilter(self)
        self.ui.molVolBox.installEventFilter(self)
        self.ui.molResultBox.installEventFilter(self)
        self.ui.dilC1Box.installEventFilter(self)
        self.ui.dilC2Box.installEventFilter(self)
        self.ui.dilV2Box.installEventFilter(self)
        self.ui.dilResultBox.installEventFilter(self)

        self.ui.molNameLine.setFocus(True)

    def _updateFormula(self, formula):
        if self.ui.displayFormulaBox.isChecked():
            self.ui.statusLabel.setText(formula)




#!/usr/bin/python3
from PyQt5 import QtWidgets, QtCore


class Completer(QtWidgets.QCompleter):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = parent
        self.setFilterMode(QtCore.Qt.MatchContains)
        self.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self.setCompletionMode(QtWidgets.QCompleter.PopupCompletion)
        self.installEventFilter(self)

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.KeyPress:
            if event.key() in (QtCore.Qt.Key_Down, QtCore.Qt.Key_Tab):
                self._nextCompletion()
                return True
            elif event.key() in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return):
                self.setCurrentRow(self.currentRow()-1)
                self.activated.emit(self.currentCompletion())
        return super().eventFilter(obj, event)

    def _nextCompletion(self):
        index = self.currentIndex()
        self.popup().setCurrentIndex(index)

        start = self.currentRow()
        if not self.setCurrentRow(start + 1):
            self.setCurrentRow(0)


class ConfirmPrompt(QtWidgets.QMessageBox):
    def __init__(self):
        super().__init__()
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)
        self.setStandardButtons(QtWidgets.QMessageBox.Apply | QtWidgets.QMessageBox.Cancel)
        self.setDefaultButton(QtWidgets.QMessageBox.Cancel)
        self.setIcon(QtWidgets.QMessageBox.Warning)



#!/usr/bin/python3
import json

units_val = \
{
    "molar (M)": 1,
    "millimolar (mM)": 10**3,
    "micromolar (μM)": 10**6,
    "nanomolar (nM)": 10**9,
    "picomolar (pM)": 10**12,
    "liters (L)": 1,
    "milliliters (mL)": 10**3,
    "microliters (μL)": 10**6,
}

units_str = \
{
    "molar (M)": "M",
    "millimolar (mM)": "mM",
    "micromolar (μM)": "μM",
    "nanomolar (nM)": "nM",
    "picomolar (pM)": "pM",
    "liters (L)": "L",
    "milliliters (mL)": "mL",
    "microliters (μL)": "μL",
}

MINUS_ONE = f"\N{SUPERSCRIPT MINUS}\N{SUPERSCRIPT ONE}"


try:
    from ..__id__ import ID
except ValueError:
    from __id__ import ID


class Calc():
    def __init__(self):
        pass

    def dilution(self, c1, c1Factor, c2, c2Factor, v2, v2Unit, resultUnit):
        try:
            c1Factor = units_val[c1Factor]
            c2Factor = units_val[c2Factor]
            v2Factor = units_val[v2Unit]
            resultFactor = units_val[resultUnit]

            # v1 = (c2 * v2) / c1
            v1 = (((c2 / c2Factor) * v2) / (c1 / c1Factor)) * (resultFactor/v2Factor)
            return v1
        except ValueError:
            return 0
        except ZeroDivisionError:
            return 0

    def dilution_string(self, c1, c1Factor, c2, c2Factor, v2, v2Unit, resultUnit):
        if c1 and c2 and v2:
            c1Unit = units_str[c1Factor]
            c2Unit = units_str[c2Factor]
            c1Factor = units_val[c1Factor]
            c2Factor = units_val[c2Factor]
            v2Factor = units_val[v2Unit]
            resultFactor = units_val[resultUnit]
            v2UnitStr = units_str[v2Unit]
            resultUnitStr = units_str[resultUnit]

            if c2Unit == "M":
                term_1st = f"({c2}{c2Unit} * {v2}{v2UnitStr})"
            else:
                term_1st = f"({c2}{c2Unit} / {c2Factor}{c2Unit}∙M{MINUS_ONE}) * {v2}{v2UnitStr})"

            if c1Unit == "M":
                term_2nd = f"{c1}{c1Unit}"
            else:
                term_2nd = f"({c1}{c1Unit} / {c1Factor}{c1Unit}∙M{MINUS_ONE})"

            if v2Factor == resultFactor:
                vol_str = f"({term_1st} / {term_2nd}"
            else:
                term_3rd = f"({resultFactor}{resultUnitStr} / {v2Factor}{v2UnitStr})"
                vol_str = f"(({term_1st} / {term_2nd}) * {term_3rd}"

            return vol_str
        return ""

    def molar_solution(self, mw, conc, concFactor, vol, volUnit):
        try:
            concFactor = units_val[concFactor]
            volUnit = units_val[volUnit]
            mass = mw * (conc / concFactor) * (vol / volUnit)
            return mass
        except ValueError:
            return 0

    def molar_string(self, mw, conc, concFactor, vol, volUnit):
        if mw and conc and vol:
            concUnit = units_str[concFactor]
            volUnitStr = units_str[volUnit]
            concFactor = units_val[concFactor]
            volUnit = units_val[volUnit]

            if concUnit == "M":
                term_1st = f"{conc}{concUnit}"
            else:
                term_1st = f"({conc}{concUnit} / {concFactor}{concUnit}∙M{MINUS_ONE})"

            if volUnitStr == "L":
                term_2nd = f"{vol}{volUnitStr}"
            else:
                term_2nd = f"({vol}{volUnitStr} / {volUnit}{volUnitStr}∙L{MINUS_ONE})"

            mass_str = f"{mw}g∙mol{MINUS_ONE} * {term_1st} * {term_2nd}"
            return mass_str
        return ""



#!/usr/bin/python3
from PyQt5 import QtCore, QtDBus

try:
    from ..__id__ import ID
except ValueError:
    from __id__ import ID


def parse(sysArgv):
    # Translate sys.argv arguments list to a dict of commands and arguments
    cmd, parsed = None, {}
    for arg in sysArgv:
        if arg.startswith("--"):
            cmd = arg[2:]
            parsed.setdefault(cmd, [])
        elif cmd:
            parsed[cmd].append(arg)
    return parsed


class QDBusObject(QtCore.QObject):
    def __init__(self, parent):
        QtCore.QObject.__init__(self)
        self.__dbusAdaptor = QDBusServerAdapter(self, parent)
        self.start()

    def start(self):
        bus = QtDBus.QDBusConnection.sessionBus()
        bus.registerObject(f"/org/{ID}", self)
        bus.registerService(f"org.{ID}")
        return bus


class QDBusServerAdapter(QtDBus.QDBusAbstractAdaptor):
    QtCore.Q_CLASSINFO("D-Bus Interface", f"org.{ID}")
    QtCore.Q_CLASSINFO("D-Bus Introspection",
    f'<interface name="org.{ID}">\n'
    '  <method name="ParseCommands">\n'
    '    <arg direction="in" type="a{s}" name="commands"/>\n'
    '  </method>\n'
    '</interface>\n')

    def __init__(self, server, parent):
        super().__init__(server)
        self.parent = parent

    @QtCore.pyqtSlot(list)
    def ParseCommands(self, sysArgv):
        cmd = QtCore.Q_ARG(dict, parse(sysArgv))
        QtCore.QMetaObject.invokeMethod(self.parent, "cli", QtCore.Qt.QueuedConnection, cmd)



#!/usr/bin/python3
import json
import math
from pathlib import Path
from itertools import combinations

try:
    from ..__id__ import ID
except ValueError:
    from __id__ import ID

SLV_FILE = Path.home() / ".config" / ID / "hsp_solvents.json"
GRP_FILE = Path.home() / ".config" / ID / "hsp_groups.json"


def combine(s1, s2, s3=None):
    if s3:
        s3_dD = s3["vol"]*s3["dD"]
        s3_dP = s3["vol"]*s3["dP"]
        s3_dH = s3["vol"]*s3["dH"]
    else:
        s3_dD, s3_dP, s3_dH = 0, 0, 0

    # dX = (a*dX1 + b*dX2) / (a+b)
    blend = {}
    blend["dD"] = (s1["vol"]*s1["dD"] + s2["vol"]*s2["dD"] + s3_dD) / 100
    blend["dP"] = (s1["vol"]*s1["dP"] + s2["vol"]*s2["dP"] + s3_dP) / 100
    blend["dH"] = (s1["vol"]*s1["dH"] + s2["vol"]*s2["dH"] + s3_dH) / 100
    return blend


def dph(target):
    with open(SLV_FILE, "r") as f:
        db = json.load(f)
    if target in db:
        return db[target]
    return None


class HSP():
    def __init__(self):
        self.abort = False
        self.reset()

    def _data_format(self, data):
        try:
            return "%.2f" % data
        except TypeError:
            pass
        return data

    def _distance(self, s1, s2, dph):
        dD, dP, dH = 0, 0, 0
        if dph[0]:
            dD = 4*(s1["dD"]-s2["dD"])**2
        if dph[1]:
            dP = (s1["dP"]-s2["dP"])**2
        if dph[2]:
            dH = (s1["dH"]-s2["dH"])**2

        ra = math.sqrt(dD + dP + dH)  # Ra^2 = {4*(dD1-dD2)^2 + (dP1-dP2)^2 +(dH1-dH2)^2}
        return round(ra, 2)

    def _single(self, target, dph):
        solvents = {}
        for name in self.available:
            if self.abort: return {}
            s = self.available[name]
            ra = self._distance(s, target, dph)
            solvents[name] = {"dD": s["dD"], "dP": s["dP"], "dH": s["dH"], "Ra": ra, "s1": name}

        for slv in solvents:
            for attr in solvents[slv]:
                value = self._data_format(solvents[slv][attr])
                solvents[slv][attr] = value
        return solvents

    def _dual(self, target, dph):
        solvents = {}
        for names in combinations(self.available, 2):
            if self.abort: return {}
            s1, s2 = names
            s1, s2 = self.available[s1], self.available[s2]
            s1["name"], s2["name"] = names
            results = {}

            for vol in range(0, 101, 1):
                s1["vol"] = 100 - vol
                s2["vol"] = 100 - s1["vol"]
                blend = combine(s1, s2)
                ra = self._distance(blend, target, dph)
                results[ra] = {"dD": blend["dD"], "dP": blend["dP"], "dH": blend["dH"], "Ra": ra,
                                    "s1": s1["name"], "s2": s2["name"], "ratio": (s1["vol"], s2["vol"])}

            best = results[min(results)]
            for key in best:
                value = self._data_format(best[key])
                best[key] = value
            key = (s1["name"], s2["name"])
            solvents[key] = best
        solvents = self._sort_dual(solvents)
        return solvents

    def _triple(self, target, dph):
        solvents = {}
        for names in combinations(self.available, 3):
            if self.abort: return {}
            s1, s2, s3 = names
            s1, s2, s3 = self.available[s1], self.available[s2], self.available[s3]
            s1["name"], s2["name"], s3["name"] = names
            results = {}

            for prop in [(i, 100-i-k, k) for i in range(1,99) for k in range(1,100-i)]:
                s1["vol"], s2["vol"], s3["vol"] = prop
                blend = combine(s1, s2, s3)
                ra = self._distance(blend, target, dph)
                results[ra] = {"dD": blend["dD"], "dP": blend["dP"], "dH": blend["dH"], "Ra": ra,
                                    "s1": s1["name"], "s2": s2["name"], "s3": s3["name"],
                                    "ratio": (s1["vol"], s2["vol"], s3["vol"])}

            best = results[min(results)]
            for key in best:
                value = self._data_format(best[key])
                best[key] = value
            key = (s1["name"], s2["name"], s3["name"])
            solvents[key] = best
        solvents = self._sort_triple(solvents)
        return solvents

    def _sort_dual(self, solvents):
        for s in dict(solvents):
            slv = solvents[s]
            r1, r2 = slv["ratio"]
            if r1 < r2:
                r1, r2 = r2, r1
                slv["s1"], slv["s2"] = slv["s2"], slv["s1"]
            slv["ratio"] = (r1, r2)

            if r1 > 98:
                del solvents[s]
        return solvents

    def _sort_triple(self, solvents):
        for s in dict(solvents):
            slv = solvents[s]
            r1, r2, r3 = slv["ratio"]

            for i in range(2):
                if r1 < r2:
                    r1, r2 = r2, r1
                    slv["s1"], slv["s2"] = slv["s2"], slv["s1"]
                if r2 < r3:
                    r2, r3 = r3, r2
                    slv["s2"], slv["s3"] = slv["s3"], slv["s2"]
            slv["ratio"] = (r1, r2, r3)

            if r3 <= 2:
                del solvents[s]
        return solvents

    def _updateGroups(self, selected_groups):
        with open(SLV_FILE, "r") as f:
            db = json.load(f)

        with open(GRP_FILE, "r") as f:
            groups = json.load(f)

        self.available = {}
        for grp in selected_groups:
            for slv in groups[grp]:
                if slv in db:
                    self.available[slv] = db[slv]

    def reset(self):
        self.all, self.single, self.dual, self.triple = {}, {}, {}, {}

    def search(self, scope, dph, target, groups):
        self._updateGroups(groups)
        if scope in ("single", "all"):
            self.single = self._single(target, dph)
        if scope in ("dual", "all"):
            self.dual = self._dual(target, dph)
        if scope in ("triple", "all"):
            self.triple = self._triple(target, dph)
        self.all = {**self.single, **self.dual, **self.triple}
        self.abort = False



#!/usr/bin/python3
import os
import math
import pandas

from openpyxl import Workbook
from openpyxl.chart import Reference, Series, ScatterChart
from openpyxl.drawing.colors import ColorChoice
from openpyxl.utils import get_column_letter
from openpyxl.styles import Font


class Chart(ScatterChart):
    def __init__(self, parent, x_min=None, x_max=0, peaks=True):
        super().__init__()
        self.parent = parent
        self.labels = parent.labels
        self.ws = parent.ws
        self.__elements__ = ('scatterStyle', 'varyColors', 'ser', 'dLbls', 'axId',)
        self._setup_axis(x_min, x_max)

        if peaks:
            self.legend = None
            self.x_axis.majorGridlines = None
            self.ws_peaks = parent.create_sheet("__peaks__")
            self.ws_peaks.sheet_state = 'hidden'
        self._add_series(peaks)

    def _add_peaks(self, column):
        x_col = get_column_letter(column)
        y_col = get_column_letter(column+1)
        index, peaks = self._find_peaks(y_col)

        for i, peak in enum(index, start=0, step=2):
            x_peak = int(self.ws[f"{x_col}{peak+2}"].value)
            self.ws_peaks[f"{x_col}{i+1}"] = x_peak
            self.ws_peaks[f"{x_col}{i+2}"] = x_peak
            self.ws_peaks[f"{y_col}{i+1}"] = 0
            self.ws_peaks[f"{y_col}{i+2}"] = self.y_axis.scaling.max

            x_values = Reference(self.ws_peaks, min_col=column, min_row=i+1, max_row=i+2)
            y_values = Reference(self.ws_peaks, min_col=column+1, min_row=i+1, max_row=i+2)
            color = self.series[-1].graphicalProperties.line.solidFill.srgbClr

            series = Series(y_values, x_values)
            series.graphicalProperties.line.width = 12721*1  # Conversion of pts to EMUs
            series.graphicalProperties.line.prstDash = "sysDash"
            series.graphicalProperties.line.solidFill = ColorChoice(srgbClr=color)
            series.smooth = True
            self.series.append(series)

    def _add_series(self, peaks):
        self.colors = ["386192", "953937", "769140", "614a7d", "35849a", "c17230",
                        "4575af", "ad3b38", "8dac4c", "745995", "3f9eb7", "e58536"]

        for i in range(1, self.ws.max_column, 2):
            x_values = Reference(self.ws, min_col=i, min_row=2, max_row=self.ws.max_row)
            y_values = Reference(self.ws, min_col=i+1, min_row=1, max_row=self.ws.max_row)
            series = Series(y_values, x_values, title_from_data=True)
            series.graphicalProperties.line.width = 12721*2  # Conversion of pts to EMUs
            series.graphicalProperties.line.solidFill = ColorChoice(srgbClr=self.colors[0])
            series.smooth = True
            self.series.append(series)
            if peaks:
                self._add_peaks(i)
            self.colors.append(self.colors.pop(0))

    def _find_peaks(self, y_col):
        values = []
        for cell in self.ws[y_col]:
            if isinstance(cell.value, float):
                if not pandas.isnull(cell.value):
                    values.append(cell.value)
        values = pandas.array(values)
        index, peaks = self._filter_peaks(values)
        return index, peaks

    def _setup_axis(self, x_min=None, x_max=0, y_min=0, y_max=0):
        scale = self.Scale(self, x_min, x_max, y_min, y_max)
        self.y_axis.scaling.min = scale.y_min
        self.x_axis.scaling.min = scale.x_min
        self.y_axis.scaling.max = scale.y_max
        self.x_axis.scaling.max = scale.x_max
        self.x_axis.number_format = "0"
        self.y_axis.number_format = "0"
        if scale.y_max < 15:
            self.y_axis.majorUnit = 1

    class Scale:
        def __init__(self, parent, x_min, x_max, y_min, y_max):
            self.x_min, self.x_max, self.y_min, self.y_max = x_min, x_max, y_min, y_max
            self.ws = parent.ws

            baseSheet = parent.parent
            x_values, minimums, maximums = baseSheet.mins_maxs(1, self.ws.max_column)
            self.x_max = max(maximums) if not x_max else x_max
            self.x_min = min(minimums) if x_min is None else x_min
            if not y_max:
                self.y_max = self._y_max(x_values, x_min, x_max)

        def _rows_limits(self, x_values, x_start, x_stop):
            # Reverse start/stop mark for descending x values
            for col in x_values:
                first = x_values[col][0]
                last = x_values[col][-1]
                break
            if first > last:
                x_stop, x_start = x_start, x_stop

            rows_limits = {}
            for i in range(1, self.ws.max_column, 2):
                letter = get_column_letter(i)
                start, stop = None, None
                for cell in self.ws[letter]:
                    if start and stop: break
                    begin = x_start if x_start else min(x_values[letter])
                    end = x_stop if x_stop else max(x_values[letter])
                    begin, end = int(begin), int(end)
                    try:
                        value = int(cell.value)
                        if value == begin:
                            start = cell.row
                        elif value == end:
                            stop = cell.row
                    except ValueError:
                        pass

                start = 2 if start is None else start
                stop = cell.row if stop is None else stop
                if start > stop:
                    start, stop = stop, start
                y_letter = get_column_letter(i+1)
                rows_limits[y_letter] = {"start": start, "stop": stop}
            return rows_limits

        def _y_max(self, x_values, x_start, x_stop):
            # For each (x) columns, find the rows corresponding to (x) axis start and stop values
            rows_limits = self._rows_limits(x_values, x_start, x_stop)

            # Create a list with all the (y) values between these rows
            y_values = []
            for col in rows_limits:
                row_min = rows_limits[col]["start"]
                row_max = rows_limits[col]["stop"]
                for cell in self.ws[col]:
                    if cell.row >= row_min and cell.row <= row_max:
                        y_values.append(cell.value)

            # Find and round the highest (y) maxima
            if y_values:
                maxima = max(y_values)
                if maxima > 200:
                    maxima = int(50 * math.ceil(maxima / 50))
                elif maxima > 50:
                    maxima = int(20 * math.ceil(maxima / 20))
                else:
                    maxima = math.ceil(maxima * 1.1)
                return maxima
            return 0


class Sheet(Workbook):
    def __init__(self, files, labels):
        super().__init__()
        self.labels = labels
        self.ws = self.active
        self.ws.title = "data"
        self._load(files)
        self._format()
        self._add_header(files)

    def _add_header(self, files):
        self.header = self._get_header(files)
        self.ws.insert_rows(1)
        for cell in self.ws["1"]:
            if cell.column > self.ws.max_column: break
            else:
                h = self.header[cell.column-1]
                cell.value = h

    def _format(self):
        for row in self.ws.iter_rows():
            for cell in row:
                value = cell.value
                try:
                    value = value.replace(",", ".")
                    value = value.replace("\ufeff", "")  # Remove BOM
                    value = float(value)
                    cell.value = value
                    cell.number_format = "0.00"
                except AttributeError:
                    pass
                except ValueError:
                    pass

    def add_stats(self, x_min, x_max):
        x_values, minimums, maximums = self.mins_maxs(1, self.ws.max_column)
        x_max = max(maximums) if not x_max else x_max
        x_min = min(minimums) if not x_min else x_min

        header = self.header[1::2]
        ws_stats = self.create_sheet("stats")
        for i, h in enumerate(header):
            ws_stats[f"A{i+1}"].value = h

        ws_stats.insert_rows(1)
        ws_stats["A1"].value = self.labels["stats title"]
        ws_stats["B1"].value = self.labels["stats x"]
        ws_stats["C1"].value = self.labels["stats y"]
        ws_stats["D1"].value = "μg/μL" ##
        for cell in ("A1", "B1", "C1", "D1"):
            ws_stats[cell].font = Font(bold=True)

        for row, column in enumerate(range(1, self.ws.max_column, 2), 2):
            x_col = get_column_letter(column)
            y_col = get_column_letter(column+1)
            values = {}
            for cell in self.ws[x_col]:
                if isinstance(cell.value, float):
                    if x_max >= cell.value >= x_min:
                        y_val = self.ws[f"{y_col}{cell.row}"].value
                        values[y_val] = cell.value
            y = max(values)
            x = values[y]
            ws_stats[f"B{row}"].value = x
            ws_stats[f"C{row}"].value = y
            ws_stats[f"D{row}"].value = f"=(C{row}-0.0456)/0.7502"  # ug/uL = (λmax – 0.0456) / 0.7502 ##
            ws_stats[f"B{row}"].number_format = "0"
            ws_stats[f"C{row}"].number_format = "0.00"
            ws_stats[f"D{row}"].number_format = "0.00"

    def mins_maxs(self, col_start, col_stop):
        # Return all the values of each (x) or (y) columns
        values = {}
        for i in range(col_start, col_stop, 2):
            letter = get_column_letter(i)
            values[letter] = []
            for cell in self.ws[letter]:
                if isinstance(cell.value, float):
                    values[letter].append(cell.value)

        minimums, maximums = [], []
        for col in values:
            minimums.append(min(values[col]))
            maximums.append(max(values[col]))
        return values, minimums, maximums


def enum(xs, start=0, step=1):
    for x in xs:
        yield (start, x)
        start += step



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'qtlab/ui/peak_ui.ui'
#
# Created by: PyQt5 UI code generator 5.13.2
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(579, 449)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setContentsMargins(10, 10, 10, 10)
        self.verticalLayout.setObjectName("verticalLayout")
        self.frame_2 = QtWidgets.QFrame(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(3)
        sizePolicy.setHeightForWidth(self.frame_2.sizePolicy().hasHeightForWidth())
        self.frame_2.setSizePolicy(sizePolicy)
        self.frame_2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_2.setObjectName("frame_2")
        self.treeLayout = QtWidgets.QVBoxLayout(self.frame_2)
        self.treeLayout.setContentsMargins(0, 0, 0, 0)
        self.treeLayout.setObjectName("treeLayout")
        self.verticalLayout.addWidget(self.frame_2)
        spacerItem = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem)
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.tabWidget.sizePolicy().hasHeightForWidth())
        self.tabWidget.setSizePolicy(sizePolicy)
        self.tabWidget.setObjectName("tabWidget")
        self.tab_3 = QtWidgets.QWidget()
        self.tab_3.setObjectName("tab_3")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.tab_3)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.groupBox = QtWidgets.QGroupBox(self.tab_3)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.typeBox = QtWidgets.QComboBox(self.groupBox)
        self.typeBox.setMinimumSize(QtCore.QSize(200, 0))
        self.typeBox.setObjectName("typeBox")
        self.typeBox.addItem("")
        self.typeBox.addItem("")
        self.verticalLayout_2.addWidget(self.typeBox)
        self.autodetectBox = QtWidgets.QCheckBox(self.groupBox)
        self.autodetectBox.setObjectName("autodetectBox")
        self.verticalLayout_2.addWidget(self.autodetectBox)
        self.horizontalLayout_2.addWidget(self.groupBox)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.groupBox_4 = QtWidgets.QGroupBox(self.tab_3)
        self.groupBox_4.setObjectName("groupBox_4")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.groupBox_4)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.focusBox = QtWidgets.QCheckBox(self.groupBox_4)
        self.focusBox.setChecked(True)
        self.focusBox.setObjectName("focusBox")
        self.gridLayout_4.addWidget(self.focusBox, 0, 0, 1, 1)
        self.focusMinBox = QtWidgets.QSpinBox(self.groupBox_4)
        self.focusMinBox.setMaximum(999999999)
        self.focusMinBox.setProperty("value", 0)
        self.focusMinBox.setObjectName("focusMinBox")
        self.gridLayout_4.addWidget(self.focusMinBox, 0, 1, 1, 1)
        self.focusMaxBox = QtWidgets.QSpinBox(self.groupBox_4)
        self.focusMaxBox.setMaximum(999999999)
        self.focusMaxBox.setProperty("value", 0)
        self.focusMaxBox.setObjectName("focusMaxBox")
        self.gridLayout_4.addWidget(self.focusMaxBox, 0, 2, 1, 1)
        self.statsBox = QtWidgets.QCheckBox(self.groupBox_4)
        self.statsBox.setChecked(True)
        self.statsBox.setObjectName("statsBox")
        self.gridLayout_4.addWidget(self.statsBox, 1, 0, 1, 1)
        self.statsMinBox = QtWidgets.QSpinBox(self.groupBox_4)
        self.statsMinBox.setMaximum(999999999)
        self.statsMinBox.setProperty("value", 0)
        self.statsMinBox.setObjectName("statsMinBox")
        self.gridLayout_4.addWidget(self.statsMinBox, 1, 1, 1, 1)
        self.statsMaxBox = QtWidgets.QSpinBox(self.groupBox_4)
        self.statsMaxBox.setMaximum(999999999)
        self.statsMaxBox.setProperty("value", 0)
        self.statsMaxBox.setObjectName("statsMaxBox")
        self.gridLayout_4.addWidget(self.statsMaxBox, 1, 2, 1, 1)
        self.horizontalLayout_2.addWidget(self.groupBox_4)
        self.tabWidget.addTab(self.tab_3, "")
        self.tab_4 = QtWidgets.QWidget()
        self.tab_4.setObjectName("tab_4")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.tab_4)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.groupBox_2 = QtWidgets.QGroupBox(self.tab_4)
        self.groupBox_2.setObjectName("groupBox_2")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBox_2)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label_2 = QtWidgets.QLabel(self.groupBox_2)
        self.label_2.setObjectName("label_2")
        self.gridLayout_2.addWidget(self.label_2, 1, 0, 1, 1)
        self.yLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.yLine.setObjectName("yLine")
        self.gridLayout_2.addWidget(self.yLine, 1, 1, 1, 1)
        self.label = QtWidgets.QLabel(self.groupBox_2)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)
        self.xLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.xLine.setObjectName("xLine")
        self.gridLayout_2.addWidget(self.xLine, 0, 1, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem2, 2, 0, 1, 1)
        self.horizontalLayout_3.addWidget(self.groupBox_2)
        self.groupBox_3 = QtWidgets.QGroupBox(self.tab_4)
        self.groupBox_3.setObjectName("groupBox_3")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBox_3)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.yStatsLine = QtWidgets.QLineEdit(self.groupBox_3)
        self.yStatsLine.setObjectName("yStatsLine")
        self.gridLayout_3.addWidget(self.yStatsLine, 2, 1, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.groupBox_3)
        self.label_4.setObjectName("label_4")
        self.gridLayout_3.addWidget(self.label_4, 2, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.groupBox_3)
        self.label_3.setObjectName("label_3")
        self.gridLayout_3.addWidget(self.label_3, 1, 0, 1, 1)
        self.xStatsLine = QtWidgets.QLineEdit(self.groupBox_3)
        self.xStatsLine.setObjectName("xStatsLine")
        self.gridLayout_3.addWidget(self.xStatsLine, 1, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.groupBox_3)
        self.label_5.setObjectName("label_5")
        self.gridLayout_3.addWidget(self.label_5, 0, 0, 1, 1)
        self.titleStatsLine = QtWidgets.QLineEdit(self.groupBox_3)
        self.titleStatsLine.setObjectName("titleStatsLine")
        self.gridLayout_3.addWidget(self.titleStatsLine, 0, 1, 1, 1)
        self.horizontalLayout_3.addWidget(self.groupBox_3)
        self.tabWidget.addTab(self.tab_4, "")
        self.verticalLayout.addWidget(self.tabWidget)
        self.frame = QtWidgets.QFrame(self.centralwidget)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setObjectName("frame")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame)
        self.horizontalLayout.setContentsMargins(-1, 6, -1, -1)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.clearButton = QtWidgets.QPushButton(self.frame)
        self.clearButton.setObjectName("clearButton")
        self.horizontalLayout.addWidget(self.clearButton)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem3)
        self.quitButton = QtWidgets.QPushButton(self.frame)
        self.quitButton.setObjectName("quitButton")
        self.horizontalLayout.addWidget(self.quitButton)
        self.convertButton = QtWidgets.QPushButton(self.frame)
        self.convertButton.setObjectName("convertButton")
        self.horizontalLayout.addWidget(self.convertButton)
        self.verticalLayout.addWidget(self.frame)
        MainWindow.setCentralWidget(self.centralwidget)
        self.statusBar = QtWidgets.QStatusBar(MainWindow)
        self.statusBar.setObjectName("statusBar")
        MainWindow.setStatusBar(self.statusBar)

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        self.groupBox.setTitle(_translate("MainWindow", "Data type"))
        self.typeBox.setItemText(0, _translate("MainWindow", "Cary WinUV (v3)"))
        self.typeBox.setItemText(1, _translate("MainWindow", "OMNIC for Dispertive Raman (v8)"))
        self.autodetectBox.setText(_translate("MainWindow", "Auto-detect"))
        self.groupBox_4.setTitle(_translate("MainWindow", "Details"))
        self.focusBox.setText(_translate("MainWindow", "Focus"))
        self.statsBox.setText(_translate("MainWindow", "Stats"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_3), _translate("MainWindow", "Options"))
        self.groupBox_2.setTitle(_translate("MainWindow", "Axis"))
        self.label_2.setText(_translate("MainWindow", "Vertical"))
        self.label.setText(_translate("MainWindow", "Horizontal"))
        self.groupBox_3.setTitle(_translate("MainWindow", "Stats"))
        self.label_4.setText(_translate("MainWindow", "Vertical"))
        self.label_3.setText(_translate("MainWindow", "Horizontal"))
        self.label_5.setText(_translate("MainWindow", "Title"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_4), _translate("MainWindow", "Labels"))
        self.clearButton.setText(_translate("MainWindow", "Clear"))
        self.quitButton.setText(_translate("MainWindow", "Quit"))
        self.convertButton.setText(_translate("MainWindow", "Convert"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'qtlab/ui/hsp_ui.ui'
#
# Created by: PyQt5 UI code generator 5.13.2
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(822, 682)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_7.setContentsMargins(10, 10, 10, 10)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setFocusPolicy(QtCore.Qt.TabFocus)
        self.tabWidget.setUsesScrollButtons(False)
        self.tabWidget.setDocumentMode(True)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.tab)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.frame_6 = QtWidgets.QFrame(self.tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_6.sizePolicy().hasHeightForWidth())
        self.frame_6.setSizePolicy(sizePolicy)
        self.frame_6.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_6.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_6.setObjectName("frame_6")
        self.searchTreeLayout = QtWidgets.QVBoxLayout(self.frame_6)
        self.searchTreeLayout.setObjectName("searchTreeLayout")
        self.horizontalLayout.addWidget(self.frame_6)
        self.frame_8 = QtWidgets.QFrame(self.tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_8.sizePolicy().hasHeightForWidth())
        self.frame_8.setSizePolicy(sizePolicy)
        self.frame_8.setMaximumSize(QtCore.QSize(280, 16777215))
        self.frame_8.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_8.setObjectName("frame_8")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.frame_8)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.groupBox_3 = QtWidgets.QGroupBox(self.frame_8)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_3.sizePolicy().hasHeightForWidth())
        self.groupBox_3.setSizePolicy(sizePolicy)
        self.groupBox_3.setObjectName("groupBox_3")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBox_3)
        self.gridLayout_3.setContentsMargins(0, -1, 0, -1)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.groupPreferedBox = QtWidgets.QCheckBox(self.groupBox_3)
        self.groupPreferedBox.setFocusPolicy(QtCore.Qt.TabFocus)
        self.groupPreferedBox.setStyleSheet("color: #004000;\n"
"background-color: #c6efce;")
        self.groupPreferedBox.setChecked(True)
        self.groupPreferedBox.setObjectName("groupPreferedBox")
        self.gridLayout_3.addWidget(self.groupPreferedBox, 0, 0, 1, 1)
        self.groupUsableBox = QtWidgets.QCheckBox(self.groupBox_3)
        self.groupUsableBox.setFocusPolicy(QtCore.Qt.TabFocus)
        self.groupUsableBox.setStyleSheet("color: #553400;\n"
"background-color: #ffeb9c;")
        self.groupUsableBox.setObjectName("groupUsableBox")
        self.gridLayout_3.addWidget(self.groupUsableBox, 0, 1, 1, 1)
        self.groupHazardousBox = QtWidgets.QCheckBox(self.groupBox_3)
        self.groupHazardousBox.setStyleSheet("color: #9c0006;\n"
"background-color: #ffc7ce;")
        self.groupHazardousBox.setObjectName("groupHazardousBox")
        self.gridLayout_3.addWidget(self.groupHazardousBox, 1, 0, 1, 1)
        self.groupCustomBox = QtWidgets.QCheckBox(self.groupBox_3)
        self.groupCustomBox.setFocusPolicy(QtCore.Qt.TabFocus)
        self.groupCustomBox.setObjectName("groupCustomBox")
        self.gridLayout_3.addWidget(self.groupCustomBox, 1, 1, 1, 1)
        self.verticalLayout_3.addWidget(self.groupBox_3)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem)
        self.groupBox = QtWidgets.QGroupBox(self.frame_8)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(4)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setContentsMargins(0, -1, 0, -1)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.frame = QtWidgets.QFrame(self.groupBox)
        self.frame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.frame)
        self.gridLayout_4.setContentsMargins(-1, 0, -1, 0)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.targetS2VolBox = QtWidgets.QSpinBox(self.frame)
        self.targetS2VolBox.setMaximum(100)
        self.targetS2VolBox.setObjectName("targetS2VolBox")
        self.gridLayout_4.addWidget(self.targetS2VolBox, 2, 1, 1, 1)
        self.targetS1VolBox = QtWidgets.QSpinBox(self.frame)
        self.targetS1VolBox.setMaximum(100)
        self.targetS1VolBox.setProperty("value", 100)
        self.targetS1VolBox.setObjectName("targetS1VolBox")
        self.gridLayout_4.addWidget(self.targetS1VolBox, 1, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.frame)
        self.label_2.setObjectName("label_2")
        self.gridLayout_4.addWidget(self.label_2, 0, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.frame)
        self.label_3.setObjectName("label_3")
        self.gridLayout_4.addWidget(self.label_3, 0, 1, 1, 1)
        self.frame_7 = QtWidgets.QFrame(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_7.sizePolicy().hasHeightForWidth())
        self.frame_7.setSizePolicy(sizePolicy)
        self.frame_7.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_7.setObjectName("frame_7")
        self.inputLayout = QtWidgets.QVBoxLayout(self.frame_7)
        self.inputLayout.setContentsMargins(0, 0, 0, 0)
        self.inputLayout.setObjectName("inputLayout")
        self.gridLayout_4.addWidget(self.frame_7, 1, 0, 2, 1)
        self.verticalLayout_2.addWidget(self.frame)
        self.frame_2 = QtWidgets.QFrame(self.groupBox)
        self.frame_2.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.gridLayout = QtWidgets.QGridLayout(self.frame_2)
        self.gridLayout.setObjectName("gridLayout")
        self.dTargetEnableBox = QtWidgets.QCheckBox(self.frame_2)
        self.dTargetEnableBox.setObjectName("dTargetEnableBox")
        self.gridLayout.addWidget(self.dTargetEnableBox, 0, 0, 1, 1)
        self.pTargetEnableBox = QtWidgets.QCheckBox(self.frame_2)
        self.pTargetEnableBox.setObjectName("pTargetEnableBox")
        self.gridLayout.addWidget(self.pTargetEnableBox, 0, 1, 1, 1)
        self.hTargetEnableBox = QtWidgets.QCheckBox(self.frame_2)
        self.hTargetEnableBox.setObjectName("hTargetEnableBox")
        self.gridLayout.addWidget(self.hTargetEnableBox, 0, 2, 1, 1)
        self.dTargetBox = QtWidgets.QDoubleSpinBox(self.frame_2)
        self.dTargetBox.setObjectName("dTargetBox")
        self.gridLayout.addWidget(self.dTargetBox, 1, 0, 1, 1)
        self.pTargetBox = QtWidgets.QDoubleSpinBox(self.frame_2)
        self.pTargetBox.setObjectName("pTargetBox")
        self.gridLayout.addWidget(self.pTargetBox, 1, 1, 1, 1)
        self.hTargetBox = QtWidgets.QDoubleSpinBox(self.frame_2)
        self.hTargetBox.setObjectName("hTargetBox")
        self.gridLayout.addWidget(self.hTargetBox, 1, 2, 1, 1)
        self.verticalLayout_2.addWidget(self.frame_2)
        self.verticalLayout_3.addWidget(self.groupBox)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem1)
        self.groupBox_4 = QtWidgets.QGroupBox(self.frame_8)
        self.groupBox_4.setObjectName("groupBox_4")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.groupBox_4)
        self.horizontalLayout_6.setContentsMargins(0, -1, 0, -1)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.searchAllButton = QtWidgets.QPushButton(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(3)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.searchAllButton.sizePolicy().hasHeightForWidth())
        self.searchAllButton.setSizePolicy(sizePolicy)
        self.searchAllButton.setObjectName("searchAllButton")
        self.horizontalLayout_6.addWidget(self.searchAllButton)
        self.searchSingleButton = QtWidgets.QPushButton(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.searchSingleButton.sizePolicy().hasHeightForWidth())
        self.searchSingleButton.setSizePolicy(sizePolicy)
        self.searchSingleButton.setObjectName("searchSingleButton")
        self.horizontalLayout_6.addWidget(self.searchSingleButton)
        self.searchDualButton = QtWidgets.QPushButton(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.searchDualButton.sizePolicy().hasHeightForWidth())
        self.searchDualButton.setSizePolicy(sizePolicy)
        self.searchDualButton.setObjectName("searchDualButton")
        self.horizontalLayout_6.addWidget(self.searchDualButton)
        self.searchTripleButton = QtWidgets.QPushButton(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.searchTripleButton.sizePolicy().hasHeightForWidth())
        self.searchTripleButton.setSizePolicy(sizePolicy)
        self.searchTripleButton.setObjectName("searchTripleButton")
        self.horizontalLayout_6.addWidget(self.searchTripleButton)
        self.verticalLayout_3.addWidget(self.groupBox_4)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem2)
        self.groupBox_2 = QtWidgets.QGroupBox(self.frame_8)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(3)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setObjectName("groupBox_2")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBox_2)
        self.gridLayout_2.setContentsMargins(0, -1, 0, -1)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.pSelectLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.pSelectLine.setAlignment(QtCore.Qt.AlignCenter)
        self.pSelectLine.setReadOnly(True)
        self.pSelectLine.setObjectName("pSelectLine")
        self.gridLayout_2.addWidget(self.pSelectLine, 2, 2, 1, 1)
        self.label_7 = QtWidgets.QLabel(self.groupBox_2)
        self.label_7.setObjectName("label_7")
        self.gridLayout_2.addWidget(self.label_7, 2, 0, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.groupBox_2)
        self.label_4.setAlignment(QtCore.Qt.AlignCenter)
        self.label_4.setObjectName("label_4")
        self.gridLayout_2.addWidget(self.label_4, 0, 2, 1, 1)
        self.dTargetLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.dTargetLine.setAlignment(QtCore.Qt.AlignCenter)
        self.dTargetLine.setReadOnly(True)
        self.dTargetLine.setObjectName("dTargetLine")
        self.gridLayout_2.addWidget(self.dTargetLine, 3, 1, 1, 1)
        self.pDeltaLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.pDeltaLine.setAlignment(QtCore.Qt.AlignCenter)
        self.pDeltaLine.setReadOnly(True)
        self.pDeltaLine.setObjectName("pDeltaLine")
        self.gridLayout_2.addWidget(self.pDeltaLine, 4, 2, 1, 1)
        self.pTargetLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.pTargetLine.setAlignment(QtCore.Qt.AlignCenter)
        self.pTargetLine.setReadOnly(True)
        self.pTargetLine.setObjectName("pTargetLine")
        self.gridLayout_2.addWidget(self.pTargetLine, 3, 2, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.groupBox_2)
        self.label_5.setAlignment(QtCore.Qt.AlignCenter)
        self.label_5.setObjectName("label_5")
        self.gridLayout_2.addWidget(self.label_5, 0, 3, 1, 1)
        self.label_8 = QtWidgets.QLabel(self.groupBox_2)
        self.label_8.setObjectName("label_8")
        self.gridLayout_2.addWidget(self.label_8, 4, 0, 1, 1)
        self.hSelectLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.hSelectLine.setAlignment(QtCore.Qt.AlignCenter)
        self.hSelectLine.setReadOnly(True)
        self.hSelectLine.setObjectName("hSelectLine")
        self.gridLayout_2.addWidget(self.hSelectLine, 2, 3, 1, 1)
        self.hDeltaLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.hDeltaLine.setAlignment(QtCore.Qt.AlignCenter)
        self.hDeltaLine.setReadOnly(True)
        self.hDeltaLine.setObjectName("hDeltaLine")
        self.gridLayout_2.addWidget(self.hDeltaLine, 4, 3, 1, 1)
        self.dDeltaLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.dDeltaLine.setAlignment(QtCore.Qt.AlignCenter)
        self.dDeltaLine.setReadOnly(True)
        self.dDeltaLine.setObjectName("dDeltaLine")
        self.gridLayout_2.addWidget(self.dDeltaLine, 4, 1, 1, 1)
        self.dSelectLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.dSelectLine.setAlignment(QtCore.Qt.AlignCenter)
        self.dSelectLine.setReadOnly(True)
        self.dSelectLine.setObjectName("dSelectLine")
        self.gridLayout_2.addWidget(self.dSelectLine, 2, 1, 1, 1)
        self.label_9 = QtWidgets.QLabel(self.groupBox_2)
        self.label_9.setObjectName("label_9")
        self.gridLayout_2.addWidget(self.label_9, 3, 0, 1, 1)
        self.label_6 = QtWidgets.QLabel(self.groupBox_2)
        self.label_6.setAlignment(QtCore.Qt.AlignCenter)
        self.label_6.setObjectName("label_6")
        self.gridLayout_2.addWidget(self.label_6, 0, 1, 1, 1)
        self.hTargetLine = QtWidgets.QLineEdit(self.groupBox_2)
        self.hTargetLine.setAlignment(QtCore.Qt.AlignCenter)
        self.hTargetLine.setReadOnly(True)
        self.hTargetLine.setObjectName("hTargetLine")
        self.gridLayout_2.addWidget(self.hTargetLine, 3, 3, 1, 1)
        self.verticalLayout_3.addWidget(self.groupBox_2)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem3)
        self.groupBox_5 = QtWidgets.QGroupBox(self.frame_8)
        self.groupBox_5.setObjectName("groupBox_5")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.groupBox_5)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.filterAllBox = QtWidgets.QRadioButton(self.groupBox_5)
        self.filterAllBox.setChecked(True)
        self.filterAllBox.setObjectName("filterAllBox")
        self.horizontalLayout_5.addWidget(self.filterAllBox)
        self.filterSingleBox = QtWidgets.QRadioButton(self.groupBox_5)
        self.filterSingleBox.setObjectName("filterSingleBox")
        self.horizontalLayout_5.addWidget(self.filterSingleBox)
        self.filterDualBox = QtWidgets.QRadioButton(self.groupBox_5)
        self.filterDualBox.setObjectName("filterDualBox")
        self.horizontalLayout_5.addWidget(self.filterDualBox)
        self.filterTripleBox = QtWidgets.QRadioButton(self.groupBox_5)
        self.filterTripleBox.setObjectName("filterTripleBox")
        self.horizontalLayout_5.addWidget(self.filterTripleBox)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem4)
        self.verticalLayout_3.addWidget(self.groupBox_5)
        self.horizontalLayout.addWidget(self.frame_8)
        self.tabWidget.addTab(self.tab, "")
        self.tab_3 = QtWidgets.QWidget()
        self.tab_3.setObjectName("tab_3")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.tab_3)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.frame_3 = QtWidgets.QFrame(self.tab_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_3.sizePolicy().hasHeightForWidth())
        self.frame_3.setSizePolicy(sizePolicy)
        self.frame_3.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_3.setObjectName("frame_3")
        self.groupTreeLayout = QtWidgets.QVBoxLayout(self.frame_3)
        self.groupTreeLayout.setObjectName("groupTreeLayout")
        self.horizontalLayout_3.addWidget(self.frame_3)
        self.frame_9 = QtWidgets.QFrame(self.tab_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_9.sizePolicy().hasHeightForWidth())
        self.frame_9.setSizePolicy(sizePolicy)
        self.frame_9.setMaximumSize(QtCore.QSize(280, 16777215))
        self.frame_9.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_9.setObjectName("frame_9")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frame_9)
        self.verticalLayout.setContentsMargins(-1, -1, -1, 6)
        self.verticalLayout.setObjectName("verticalLayout")
        self.groupBox_6 = QtWidgets.QGroupBox(self.frame_9)
        self.groupBox_6.setObjectName("groupBox_6")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.groupBox_6)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.groupPreferedButton = QtWidgets.QRadioButton(self.groupBox_6)
        self.groupPreferedButton.setStyleSheet("color: #004000;\n"
"background-color: #c6efce;")
        self.groupPreferedButton.setChecked(True)
        self.groupPreferedButton.setObjectName("groupPreferedButton")
        self.gridLayout_6.addWidget(self.groupPreferedButton, 0, 0, 1, 1)
        self.groupUsableButton = QtWidgets.QRadioButton(self.groupBox_6)
        self.groupUsableButton.setStyleSheet("color: #553400;\n"
"background-color: #ffeb9c;")
        self.groupUsableButton.setObjectName("groupUsableButton")
        self.gridLayout_6.addWidget(self.groupUsableButton, 0, 1, 1, 1)
        self.groupHazardousButton = QtWidgets.QRadioButton(self.groupBox_6)
        self.groupHazardousButton.setStyleSheet("color: #9c0006;\n"
"background-color: #ffc7ce;")
        self.groupHazardousButton.setObjectName("groupHazardousButton")
        self.gridLayout_6.addWidget(self.groupHazardousButton, 1, 0, 1, 1)
        self.groupCustomButton = QtWidgets.QRadioButton(self.groupBox_6)
        self.groupCustomButton.setObjectName("groupCustomButton")
        self.gridLayout_6.addWidget(self.groupCustomButton, 1, 1, 1, 1)
        self.verticalLayout.addWidget(self.groupBox_6)
        spacerItem5 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem5)
        self.groupBox_7 = QtWidgets.QGroupBox(self.frame_9)
        self.groupBox_7.setObjectName("groupBox_7")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.groupBox_7)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.label = QtWidgets.QLabel(self.groupBox_7)
        self.label.setObjectName("label")
        self.gridLayout_5.addWidget(self.label, 4, 0, 1, 1)
        self.groupNameLine = QtWidgets.QLineEdit(self.groupBox_7)
        self.groupNameLine.setObjectName("groupNameLine")
        self.gridLayout_5.addWidget(self.groupNameLine, 1, 0, 1, 7)
        self.label_11 = QtWidgets.QLabel(self.groupBox_7)
        self.label_11.setObjectName("label_11")
        self.gridLayout_5.addWidget(self.label_11, 3, 0, 1, 1)
        self.hGroupBox = QtWidgets.QDoubleSpinBox(self.groupBox_7)
        self.hGroupBox.setObjectName("hGroupBox")
        self.gridLayout_5.addWidget(self.hGroupBox, 4, 1, 1, 1)
        self.dGroupBox = QtWidgets.QDoubleSpinBox(self.groupBox_7)
        self.dGroupBox.setObjectName("dGroupBox")
        self.gridLayout_5.addWidget(self.dGroupBox, 3, 1, 1, 1)
        self.pGroupBox = QtWidgets.QDoubleSpinBox(self.groupBox_7)
        self.pGroupBox.setObjectName("pGroupBox")
        self.gridLayout_5.addWidget(self.pGroupBox, 3, 4, 1, 1)
        self.mvGroupBox = QtWidgets.QDoubleSpinBox(self.groupBox_7)
        self.mvGroupBox.setObjectName("mvGroupBox")
        self.gridLayout_5.addWidget(self.mvGroupBox, 4, 4, 1, 1)
        spacerItem6 = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.gridLayout_5.addItem(spacerItem6, 6, 0, 1, 1)
        self.label_10 = QtWidgets.QLabel(self.groupBox_7)
        self.label_10.setObjectName("label_10")
        self.gridLayout_5.addWidget(self.label_10, 3, 3, 1, 1)
        self.label_12 = QtWidgets.QLabel(self.groupBox_7)
        self.label_12.setObjectName("label_12")
        self.gridLayout_5.addWidget(self.label_12, 4, 3, 1, 1)
        self.groupAddButton = QtWidgets.QPushButton(self.groupBox_7)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupAddButton.sizePolicy().hasHeightForWidth())
        self.groupAddButton.setSizePolicy(sizePolicy)
        self.groupAddButton.setMaximumSize(QtCore.QSize(160, 16777215))
        self.groupAddButton.setObjectName("groupAddButton")
        self.gridLayout_5.addWidget(self.groupAddButton, 7, 0, 1, 2)
        self.groupClearButton = QtWidgets.QPushButton(self.groupBox_7)
        self.groupClearButton.setObjectName("groupClearButton")
        self.gridLayout_5.addWidget(self.groupClearButton, 7, 3, 1, 2)
        self.verticalLayout.addWidget(self.groupBox_7)
        spacerItem7 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem7)
        self.frame_4 = QtWidgets.QFrame(self.frame_9)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.frame_4.sizePolicy().hasHeightForWidth())
        self.frame_4.setSizePolicy(sizePolicy)
        self.frame_4.setObjectName("frame_4")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.frame_4)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem8 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem8)
        self.resetButton = QtWidgets.QPushButton(self.frame_4)
        self.resetButton.setObjectName("resetButton")
        self.horizontalLayout_2.addWidget(self.resetButton)
        self.verticalLayout.addWidget(self.frame_4)
        self.horizontalLayout_3.addWidget(self.frame_9)
        self.tabWidget.addTab(self.tab_3, "")
        self.verticalLayout_7.addWidget(self.tabWidget)
        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)
        MainWindow.setTabOrder(self.tabWidget, self.dSelectLine)
        MainWindow.setTabOrder(self.dSelectLine, self.pSelectLine)
        MainWindow.setTabOrder(self.pSelectLine, self.hSelectLine)
        MainWindow.setTabOrder(self.hSelectLine, self.dDeltaLine)
        MainWindow.setTabOrder(self.dDeltaLine, self.pDeltaLine)
        MainWindow.setTabOrder(self.pDeltaLine, self.hDeltaLine)
        MainWindow.setTabOrder(self.hDeltaLine, self.groupPreferedBox)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.groupBox_3.setTitle(_translate("MainWindow", "Solvents groups"))
        self.groupPreferedBox.setText(_translate("MainWindow", "Preferred"))
        self.groupUsableBox.setText(_translate("MainWindow", "Usable"))
        self.groupHazardousBox.setText(_translate("MainWindow", "Hazardous"))
        self.groupCustomBox.setText(_translate("MainWindow", "Custom"))
        self.groupBox.setTitle(_translate("MainWindow", "Target"))
        self.label_2.setText(_translate("MainWindow", "Solvent(s)"))
        self.label_3.setText(_translate("MainWindow", "% Vol."))
        self.dTargetEnableBox.setText(_translate("MainWindow", "𝛿D"))
        self.pTargetEnableBox.setText(_translate("MainWindow", "𝛿P"))
        self.hTargetEnableBox.setText(_translate("MainWindow", "𝛿H"))
        self.groupBox_4.setTitle(_translate("MainWindow", "Search"))
        self.searchAllButton.setText(_translate("MainWindow", "Optimize"))
        self.searchSingleButton.setText(_translate("MainWindow", "1"))
        self.searchDualButton.setText(_translate("MainWindow", "2"))
        self.searchTripleButton.setText(_translate("MainWindow", "3"))
        self.groupBox_2.setTitle(_translate("MainWindow", "Selection"))
        self.label_7.setText(_translate("MainWindow", "Current"))
        self.label_4.setText(_translate("MainWindow", "𝛿P"))
        self.label_5.setText(_translate("MainWindow", "𝛿H"))
        self.label_8.setText(_translate("MainWindow", "Delta"))
        self.label_9.setText(_translate("MainWindow", "Target"))
        self.label_6.setText(_translate("MainWindow", "𝛿D"))
        self.groupBox_5.setTitle(_translate("MainWindow", "Filter view"))
        self.filterAllBox.setText(_translate("MainWindow", "All"))
        self.filterSingleBox.setText(_translate("MainWindow", "Single"))
        self.filterDualBox.setText(_translate("MainWindow", "Dual"))
        self.filterTripleBox.setText(_translate("MainWindow", "Triple"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("MainWindow", "Search results"))
        self.groupBox_6.setTitle(_translate("MainWindow", "Current Group"))
        self.groupPreferedButton.setText(_translate("MainWindow", "Preferred"))
        self.groupUsableButton.setText(_translate("MainWindow", "Usable"))
        self.groupHazardousButton.setText(_translate("MainWindow", "Hazardous"))
        self.groupCustomButton.setText(_translate("MainWindow", "Custom"))
        self.groupBox_7.setTitle(_translate("MainWindow", "New Solvent"))
        self.label.setText(_translate("MainWindow", "𝛿H"))
        self.groupNameLine.setPlaceholderText(_translate("MainWindow", "Name"))
        self.label_11.setText(_translate("MainWindow", "𝛿D"))
        self.label_10.setText(_translate("MainWindow", "𝛿P"))
        self.label_12.setText(_translate("MainWindow", "MV"))
        self.groupAddButton.setText(_translate("MainWindow", "Add"))
        self.groupClearButton.setText(_translate("MainWindow", "Clear"))
        self.resetButton.setText(_translate("MainWindow", "Reset to default"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_3), _translate("MainWindow", "Manage groups"))



#!/usr/bin/python3
import sys
from PyQt5 import QtWidgets, QtCore

try:
    from .__id__ import ID, APP_NAME
    from .backend import cli
except ImportError:
    from __id__ import ID, APP_NAME
    from backend import cli


class Main(QtCore.QObject):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.clipboard = parent.clipboard
        self.calc = QtWidgets.QWidget()
        self.hsp = QtWidgets.QWidget()
        self.peak = QtWidgets.QWidget()

    def _calcImport(self):
        try:
            from .frontend import calc
        except ImportError:
            from frontend import calc
        self.calc = calc.Main(self)

    def _hspImport(self):
        try:
            from .frontend import hsp
        except ImportError:
            from frontend import hsp
        self.hsp = hsp.Main(self)

    def _peakImport(self):
        try:
            from .frontend import peak
        except ImportError:
            from frontend import peak
        self.peak = peak.Main(self)

    @QtCore.pyqtSlot(dict)
    def cli(self, cmd):
        if "quit" in cmd:
            sys.exit()

        elif "hsp" in cmd:
            if not self.hsp.isVisible():
                self._hspImport()

        elif "peak" in cmd:
            if not self.peak.isVisible():
                self._peakImport()

        elif "calc" in cmd:
            if "toggle" in cmd and self.calc.isVisible():
                self.calc.close()
            else:
                self._calcImport()


def main(cmd=""):
    app = QtWidgets.QApplication(sys.argv)
    app.setDesktopFileName(ID)
    app.setApplicationName(APP_NAME)
    cmd = cli.parse(cmd)
    widget = Main(app)
    widget.cli(cmd)
    widgetBus = cli.QDBusObject(parent=widget)
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



#!/usr/bin/python3
import copy
import json

from pathlib import Path
from PyQt5 import QtWidgets

try:
    from ..__db__ import DEFAULT
except ValueError:
    from __db__ import DEFAULT


class Database():
    def __init__(self, path):
        self.path = path
        path = Path(path)
        self.name = path.stem
        if path.is_file() and path.stat().st_size > 0:
            self._load()
            self._validate()
        else:
            path.parents[0].mkdir(parents=True, exist_ok=True)
            self.db = copy.deepcopy(DEFAULT)
            with open(path, "w") as f:
                f.write(json.dumps(self.db, indent=2, sort_keys=False))
            print(f"Created preferences file for '{self.name}'")

    def _load(self):
        with open(self.path, "r") as f:
            self.db = json.load(f)
        print(f"Loaded {self.name} database")

    def _repair(self, db, default=DEFAULT):
        for key in default:
            if isinstance(default[key], dict):
                db.setdefault(key, default[key])
                self._repair(db[key], default[key])
            else:
                db.setdefault(key, default[key])
            if not isinstance(default[key], type(db[key])):
                print(f"Fixed type for '{key}': {type(key)} to {type(default[key])}")
                db[key] = default.get(key)

    def _save(self):
        with open(self.path, "w") as f:
            f.write(json.dumps(self.db, indent=2, sort_keys=False))
        print(f"Saved {self.name} database")

    def _validate(self):
        old = json.dumps(self.db, sort_keys=True)
        self._repair(self.db)
        new = json.dumps(self.db, sort_keys=True)
        if not new == old:
            print(f"Repaired missing keys in {self.name} database")
            self._save()

    def load(self, db):
        for obj in self.fields:
            value = db[obj.objectName()]
            if isinstance(obj, QtWidgets.QLineEdit):
                obj.setText(value)
            elif isinstance(obj, QtWidgets.QComboBox):
                obj.setCurrentText(value)
            elif isinstance(obj, QtWidgets.QTextEdit) or isinstance(obj, QtWidgets.QPlainTextEdit):
                obj.setPlainText(value)
            elif isinstance(obj, QtWidgets.QCheckBox) or isinstance(obj, QtWidgets.QRadioButton):
                obj.setChecked(value)
            elif isinstance(obj, QtWidgets.QSpinBox) or isinstance(obj, QtWidgets.QDoubleSpinBox):
                obj.setValue(value)
            else:
                print(f"Could not handle object type {type(obj)}")

    def save(self, db):
        for obj in self.fields:
            name = obj.objectName()
            if isinstance(obj, QtWidgets.QLineEdit):
                db[name] = obj.text()
            elif isinstance(obj, QtWidgets.QComboBox):
                db[name] = obj.currentText()
            elif isinstance(obj, QtWidgets.QTextEdit) or isinstance(obj, QtWidgets.QPlainTextEdit):
                db[name] = obj.toPlainText()
            elif isinstance(obj, QtWidgets.QCheckBox) or isinstance(obj, QtWidgets.QRadioButton):
                db[name] = obj.isChecked()
            elif isinstance(obj, QtWidgets.QSpinBox) or isinstance(obj, QtWidgets.QDoubleSpinBox):
                db[name] = obj.value()
        self._save()



#!/usr/bin/python3
import pandas
import scipy.signal

from pathlib import Path
from openpyxl.utils import get_column_letter
from openpyxl.utils.dataframe import dataframe_to_rows

try:
    from ..backend import peak_base
except ValueError:
    from backend import peak_base


class RamanChart(peak_base.Chart):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x_axis.title = self.labels["x"]
        self.y_axis.title = self.labels["y"]

    def _filter_peaks(self, values):
        average = sum(values) / len(values)
        prominence = int(average * 0.25)
        index, peaks = scipy.signal.find_peaks(values, prominence=prominence)
        return index, peaks


class RamanSheet(peak_base.Sheet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _get_header(self, files):
        header = []
        for f in files:
            f = Path(f)
            if f.is_file() and f.suffix.lower() == ".csv":
                header += [self.labels["x"], f.stem]
        return header

    def _load(self, files):
        combined_csv = pandas.concat([pandas.read_csv(f, sep=",", encoding="latin-1") for f in files], axis=1)
        combined_csv = combined_csv.dropna(axis='columns', how='all')  # Drop empty columns
        for row in dataframe_to_rows(combined_csv, index=False, header=False):
            self.ws.append(row)

    def add_chart(self, pos, **kwargs):
        chart = RamanChart(self, **kwargs)
        self.ws.add_chart(chart, pos)


class UVChart(peak_base.Chart):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x_axis.title = self.labels["x"]
        self.y_axis.title = self.labels["y"]

    def _filter_peaks(self, values):
        index, peaks = scipy.signal.find_peaks(values, prominence=0.01, width=3)
        return index, peaks


class UVSheet(peak_base.Sheet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ws.delete_rows(2, 3)

    def _get_header(self, files=None):
        header = []
        for i in range(1, self.ws.max_column, 2):
            letter = get_column_letter(i)
            header.append(self.labels["x"])
            header.append(self.ws[f"{letter}1"].value)
        return header

    def _load(self, files):
        combined_csv = pandas.concat([pandas.read_csv(f, sep=";", encoding="latin-1") for f in files], axis=1)
        combined_csv = combined_csv.dropna(axis='columns', how='all')  # Drop empty columns
        for row in dataframe_to_rows(combined_csv, index=False):
            self.ws.append(row)

    def add_chart(self, pos, **kwargs):
        chart = UVChart(self, **kwargs)
        self.ws.add_chart(chart, pos)
